# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Curved polygon and associated helpers.

A curved polygon (in :math:`\mathbf{R}^2`) is defined by the
collection of B |eacute| zier curves that determine the
boundary.

.. |eacute| unicode:: U+000E9 .. LATIN SMALL LETTER E WITH ACUTE
   :trim:

.. testsetup:: *

   import numpy as np
   import bezier
"""

from bezier import _helpers
from bezier import _plot_helpers
from bezier import _triangle_helpers


class CurvedPolygon:
    """Represents an object defined by its curved boundary.

    The boundary is a piecewise defined collection of
    B |eacute| zier curves.

    .. note::

       The direction of the nodes in each :class:`.Curve`
       on the boundary is important. When verifying, we check that
       one curve begins where the last one ended.

    .. image:: ../../images/curved_polygon_constructor1.png
       :align: center

    .. doctest:: curved-polygon-constructor

       >>> import bezier
       >>> nodes0 = np.asfortranarray([
       ...     [0.0,  1.0, 2.0],
       ...     [0.0, -1.0, 0.0],
       ... ])
       >>> edge0 = bezier.Curve(nodes0, degree=2)
       >>> nodes1 = np.asfortranarray([
       ...     [2.0, 2.0],
       ...     [0.0, 1.0],
       ... ])
       >>> edge1 = bezier.Curve(nodes1, degree=1)
       >>> nodes2 = np.asfortranarray([
       ...     [2.0, 1.0, 0.0],
       ...     [1.0, 2.0, 1.0],
       ... ])
       >>> edge2 = bezier.Curve(nodes2, degree=2)
       >>> nodes3 = np.asfortranarray([
       ...     [0.0, 0.0],
       ...     [1.0, 0.0],
       ... ])
       >>> edge3 = bezier.Curve(nodes3, degree=1)
       >>> curved_poly = bezier.CurvedPolygon(
       ...     edge0, edge1, edge2, edge3)
       >>> curved_poly
       <CurvedPolygon (num_sides=4)>

    .. testcleanup:: curved-polygon-constructor

       import make_images
       make_images.curved_polygon_constructor1(curved_poly)

    Though the endpoints of each pair of edges are verified to match,
    the curved polygon as a whole is not verified, so creating
    a curved polygon with self-intersections is possible:

    .. image:: ../../images/curved_polygon_constructor2.png
       :align: center

    .. doctest:: curved-polygon-constructor-invalid

       >>> nodes0 = np.asfortranarray([
       ...     [0.0, 1.0],
       ...     [0.0, 0.0],
       ... ])
       >>> edge0 = bezier.Curve(nodes0, degree=1)
       >>> nodes1 = np.asfortranarray([
       ...     [1.0, 1.25, 1.0],
       ...     [0.0, 0.5 , 1.0],
       ... ])
       >>> edge1 = bezier.Curve(nodes1, degree=2)
       >>> nodes2 = np.asfortranarray([
       ...     [1.0, 2.0],
       ...     [1.0, 1.0],
       ... ])
       >>> edge2 = bezier.Curve(nodes2, degree=1)
       >>> nodes3 = np.asfortranarray([
       ...     [2.0, 1.0 , 0.0],
       ...     [1.0, 0.75, 0.0]
       ... ])
       >>> edge3 = bezier.Curve(nodes3, degree=2)
       >>> curved_poly = bezier.CurvedPolygon(
       ...     edge0, edge1, edge2, edge3)
       >>> curved_poly
       <CurvedPolygon (num_sides=4)>

    .. testcleanup:: curved-polygon-constructor-invalid

       import make_images
       make_images.curved_polygon_constructor2(curved_poly)

    Args:
        edges (Tuple[~bezier.curve.Curve, ...]): The boundary edges
            of the curved polygon.
        kwargs: There are two keyword arguments accepted:

            * ``metadata`` (:class:`~typing.Sequence`): A sequence of triples
              associated with this curved polygon. This is intended to be used
              by callers that have created a curved polygon as an intersection
              between two B |eacute| zier triangles.
            * ``_verify`` (:class:`bool`): Indicates if the edges should be
              verified as having shared endpoints. Defaults to :data:`True`.

            Other keyword arguments specified will be silently ignored.
    """

    __slots__ = ("_edges", "_num_sides", "_metadata")

    def __init__(self, *edges, **kwargs):
        self._edges = edges
        self._num_sides = len(edges)
        self._metadata = kwargs.pop("metadata", None)
        if kwargs.pop("_verify", True):
            self._verify()

    @staticmethod
    def _verify_pair(prev, curr):
        """Verify a pair of sides share an endpoint.

        .. note::

           This currently checks that edge endpoints match **exactly**
           but allowing some roundoff may be desired.

        Args:
            prev (.Curve): "Previous" curve at piecewise junction.
            curr (.Curve): "Next" curve at piecewise junction.

        Raises:
            ValueError: If the previous side is not in 2D.
            ValueError: If consecutive sides don't share an endpoint.
        """
        if prev._dimension != 2:
            raise ValueError("Curve not in R^2", prev)

        end = prev._nodes[:, -1]
        start = curr._nodes[:, 0]
        if not _helpers.vector_close(end, start):
            raise ValueError(
                "Not sufficiently close",
                "Consecutive sides do not have common endpoint",
                prev,
                curr,
            )

    def _verify(self):
        """Verify that the edges define a curved polygon.

        This may not be entirely comprehensive, e.g. won't check
        self-intersection of the defined polygon.

        .. note::

           This currently checks that edge endpoints match **exactly**
           but allowing some roundoff may be desired.

        Raises:
            ValueError: If there are fewer than two sides.
            ValueError: If one of the sides is not in 2D.
            ValueError: If consecutive sides don't share an endpoint.
        """
        if self._num_sides < 2:
            raise ValueError("At least two sides required.")

        for prev, curr in zip(self._edges, self._edges[1:]):
            self._verify_pair(prev, curr)
        # Now we check that the final edge wraps around.
        prev = self._edges[-1]
        curr = self._edges[0]
        self._verify_pair(prev, curr)

    @property
    def num_sides(self):
        """int: The number of sides in the current polygon."""
        return self._num_sides

    @property
    def __dict__(self):
        """dict: Dictionary of current curved polygon's property namespace.

        This is just a stand-in property for the usual ``__dict__``. This
        class defines ``__slots__`` so by default would not provide a
        ``__dict__``.

        This also means that the current object can't be modified by the
        returned dictionary.
        """
        return {"_edges": self._edges, "_num_sides": self._num_sides}

    @property
    def area(self):
        r"""The area of the current curved polygon.

        This assumes, but does not check, that the current curved polygon
        is valid (i.e. it is bounded by the edges).

        This computes the area via Green's theorem. Using the vector field
        :math:`\mathbf{F} = \left[-y, x\right]^T`, since
        :math:`\partial_x(x) - \partial_y(-y) = 2` Green's theorem says

        .. math::

           \int_{\mathcal{P}} 2 \, d\textbf{x} =
           \int_{\partial \mathcal{P}} -y \, dx + x \, dy

        (where :math:`\mathcal{P}` is the current curved polygon).

        Note that for a given edge :math:`C(r)` with control points
        :math:`x_j, y_j`, the integral can be simplified:

        .. math::

            \int_C -y \, dx + x \, dy = \int_0^1 (x y' - y x') \, dr
                = \sum_{i < j} (x_i y_j - y_i x_j) \int_0^1 b_{i, d}
                b'_{j, d} \, dr

        where :math:`b_{i, d}, b_{j, d}` are Bernstein basis polynomials.

        Returns:
            float: The area of the current curved polygon.
        """
        edges = tuple(edge._nodes for edge in self._edges)
        return _triangle_helpers.compute_area(edges)

    def __repr__(self):
        """Representation of current object.

        Returns:
            str: Object representation.
        """
        return "<{} (num_sides={:d})>".format(
            self.__class__.__name__, self._num_sides
        )

    def plot(self, pts_per_edge, color=None, ax=None):
        """Plot the current curved polygon.

        Args:
            pts_per_edge (int): Number of points to plot per curved edge.
            color (Optional[Tuple[float, float, float]]): Color as RGB profile.
            ax (Optional[matplotlib.artist.Artist]): matplotlib axis object
                to add plot to.

        Returns:
            matplotlib.artist.Artist: The axis containing the plot. This
            may be a newly created axis.
        """
        if ax is None:
            ax = _plot_helpers.new_axis()
        _plot_helpers.add_patch(ax, color, pts_per_edge, *self._edges)
        return ax



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic geometry and floating point helpers.

The functions provided by this module have a Cython speedup with the
exact same interface which calls out to a Fortran implementation. The speedup
will be used if the extension can be built.
"""

from bezier import _py_helpers

try:
    from bezier import _speedup
except ImportError:  # pragma: NO COVER
    _speedup = None


# pylint: disable=invalid-name
if _speedup is None:  # pragma: NO COVER
    vector_close = _py_helpers.vector_close
    in_interval = _py_helpers.in_interval
    bbox = _py_helpers.bbox
    contains_nd = _py_helpers.contains_nd
    cross_product = _py_helpers.cross_product
    wiggle_interval = _py_helpers.wiggle_interval
    simple_convex_hull = _py_helpers.simple_convex_hull
    polygon_collide = _py_helpers.polygon_collide
else:
    vector_close = _speedup.vector_close
    in_interval = _speedup.in_interval
    bbox = _speedup.bbox
    contains_nd = _speedup.contains_nd
    cross_product = _speedup.cross_product
    wiggle_interval = _speedup.wiggle_interval
    simple_convex_hull = _speedup.simple_convex_hull
    polygon_collide = _speedup.polygon_collide
# pylint: enable=invalid-name



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic (i.e. platform-independent) helpers for ``setup.py``."""

import copy
import distutils.ccompiler
import os
import shutil
import subprocess
import sys

import setuptools
import setuptools.command.build_ext


DEBUG_ENV = "DEBUG"
WHEEL_ENV = "BEZIER_WHEEL"
"""Environment variable used to indicate a wheel is being built.

If this is present (e.g. ``BEZIER_WHEEL="True"``) then the ``-march=native``
flag will **not** be used.
"""
FORTRAN_LIBRARY_PREFIX = "\nlibraries: ="
GFORTRAN_MISSING_LIBS = """\
``gfortran`` default library path not found via:

$ gfortran -print-search-dirs
{}"""
GFORTRAN_BAD_PATH = "``gfortran`` library path {} is not a directory."
# NOTE: These are mostly recommendations from Certik found here:
#         http://www.fortran90.org/src/faq.html
#       specifically "What compiler options should I use for ...?".
FPIC = "-fPIC"
GFORTRAN_SHARED_FLAGS = (  # Used for both "DEBUG" and "OPTIMIZE"
    "-Wall",
    "-Wextra",
    # ``-Wextra`` includes ``no-compare-reals``, which warns about
    # ``value == 0.0_dp``
    "-Wno-compare-reals",
    "-Wimplicit-interface",
    FPIC,
    "-fmax-errors=1",
    "-std=f2008",
)
GFORTRAN_DEBUG_FLAGS = (
    "-g",
    "-fcheck=all",
    "-fbacktrace",
    "-fimplicit-none",
    "-pedantic",
)
GFORTRAN_NATIVE_FLAG = "-march=native"
GFORTRAN_OPTIMIZE_FLAGS = ("-Werror", "-O3", "-funroll-loops")
BAD_JOURNAL = "Saving journal failed with {!r}."
JOURNAL_ENV = "BEZIER_JOURNAL"
"""Environment variable to specify a text file for saving compiler commands.

Can be used to determine how the binary extension was compiled. This can be
useful, for example, to track changes across different systems or simply
to make sure the build is occurring as expected.
"""
QUADPACK_DIR = "quadpack"
QUADPACK_SOURCE_FILENAME = os.path.join("src", "fortran", QUADPACK_DIR, "{}.f")
# NOTE: QUADPACK module dependencies: order is important.
QUADPACK_MODULES = ("d1mach", "dqelg", "dqpsrt", "dqk21", "dqagse")
# NOTE: This represents the Fortran module dependency graph. Order is
#       important both of the keys and of the dependencies that are in
#       each value.
FORTRAN_MODULES = (
    "types",
    "status",
    "helpers",
    "curve",
    "triangle",
    "curve_intersection",
    "triangle_intersection",
)
FORTRAN_INCLUDE_DIR = os.path.join("src", "fortran", "include")
FORTRAN_SOURCE_FILENAME = os.path.join("src", "fortran", "{}.f90")
OBJECT_FILENAME = os.path.join("src", "fortran", "{}.o")
SPEEDUP_FILENAME = os.path.join("src", "python", "bezier", "_speedup.c")


def gfortran_search_path(library_dirs):
    """Get the library directory paths for ``gfortran``.

    Looks for ``libraries: =`` in the output of ``gfortran -print-search-dirs``
    and then parses the paths. If this fails for any reason, this method will
    print an error and return ``library_dirs``.

    Args:
        library_dirs (List[str]): Existing library directories.

    Returns:
        List[str]: The library directories for ``gfortran``.
    """
    cmd = ("gfortran", "-print-search-dirs")
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    return_code = process.wait()
    # Bail out if the command failed.
    if return_code != 0:
        return library_dirs

    cmd_output = process.stdout.read().decode("utf-8")
    parts = cmd_output.split(FORTRAN_LIBRARY_PREFIX)
    if len(parts) != 2:
        msg = GFORTRAN_MISSING_LIBS.format(cmd_output)
        print(msg, file=sys.stderr)
        return library_dirs

    library_lines = parts[1].split("\n", 1)
    library_line = library_lines[0]
    # Go through each library in the ``libraries: = ...`` line.
    accepted = set(library_dirs)
    for part in library_line.split(os.pathsep):
        full_path = os.path.abspath(part.strip())
        if os.path.isdir(full_path):
            accepted.add(full_path)
        else:
            # Ignore anything that isn't a directory.
            msg = GFORTRAN_BAD_PATH.format(full_path)
            print(msg, file=sys.stderr)
    return sorted(accepted)


def extension_modules():
    import numpy as np

    libraries, library_dirs = BuildFortranThenExt.get_library_dirs()
    if BuildFortranThenExt.USE_SHARED_LIBRARY:
        # Here we don't depend on object files since the functionality
        # is contained in the shared library.
        extra_objects = []
    else:
        # NOTE: These may be treated as relative paths and replaced
        #       before the extension is actually built.
        extra_objects = [
            OBJECT_FILENAME.format(fortran_module)
            for fortran_module in FORTRAN_MODULES
        ]
        extra_objects.extend(
            OBJECT_FILENAME.format(os.path.join(QUADPACK_DIR, fortran_module))
            for fortran_module in QUADPACK_MODULES
        )
    # NOTE: Copy ``libraries`` and ``library_dirs`` so they
    #       aren't shared and mutable.
    extension = setuptools.Extension(
        "bezier._speedup",
        [SPEEDUP_FILENAME],
        extra_objects=extra_objects,
        include_dirs=[np.get_include(), FORTRAN_INCLUDE_DIR],
        libraries=copy.deepcopy(libraries),
        library_dirs=copy.deepcopy(library_dirs),
    )
    return [extension]


def _update_flags(compiler_flags, remove_flags=()):
    """Update a given set of compiler flags.

    Args:
        compiler_flags (List[str]): Existing flags associated with a compiler.
        remove_flags (Optional[Container[str]]): A container of flags to remove
            that will override any of the defaults.

    Returns:
        List[str]: The modified list (i.e. some flags added and some removed).
    """
    for flag in GFORTRAN_SHARED_FLAGS:
        if flag not in compiler_flags:
            compiler_flags.append(flag)
    if DEBUG_ENV in os.environ:
        to_add = GFORTRAN_DEBUG_FLAGS
        to_remove = GFORTRAN_OPTIMIZE_FLAGS
    else:
        to_add = GFORTRAN_OPTIMIZE_FLAGS
        if os.environ.get(WHEEL_ENV) is None:
            to_add += (GFORTRAN_NATIVE_FLAG,)
        to_remove = GFORTRAN_DEBUG_FLAGS
    for flag in to_add:
        if flag not in compiler_flags:
            compiler_flags.append(flag)
    return [
        flag
        for flag in compiler_flags
        if not (flag in to_remove or flag in remove_flags)
    ]


def patch_f90_compiler(f90_compiler):
    """Patch up ``f90_compiler``.

    For now, only updates the flags for ``gfortran``. In this case, it add
    any of ``GFORTRAN_SHARED_FLAGS`` that are missing. In debug mode, it also
    adds any flags in ``GFORTRAN_DEBUG_FLAGS`` and makes sure none of the flags
    in ``GFORTRAN_OPTIMIZE_FLAGS`` are present. In standard mode ("OPTIMIZE"),
    makes sure flags in ``GFORTRAN_OPTIMIZE_FLAGS`` are present and flags in
    ``GFORTRAN_DEBUG_FLAGS`` are not.

    Args:
        f90_compiler (numpy.distutils.fcompiler.FCompiler): A Fortran compiler
            instance.
    """
    # NOTE: NumPy may not be installed, but we don't want **this** module to
    #       cause an import failure in ``setup.py``.
    from numpy.distutils.fcompiler import gnu

    # Only ``gfortran``.
    if not isinstance(f90_compiler, gnu.Gnu95FCompiler):
        return

    f90_compiler.compiler_f77[:] = _update_flags(
        f90_compiler.compiler_f77, remove_flags=("-Werror",)
    )
    f90_compiler.compiler_f90[:] = _update_flags(f90_compiler.compiler_f90)


class BuildFortranThenExt(setuptools.command.build_ext.build_ext):
    """Custom ``build_ext`` command.

    * Builds Fortran object files for each module (these are required by
      the Cython-generated ``*.c`` extension modules)
    * Provides a static library at ``bezier/lib/libbezier.a`` (on Windows
      this is ``bezier/lib/bezier.lib``)
    * Provides an optional "journaling" feature which allows commands invoked
      during the compilation to be logged (or journaled) to a text file.

    Provides mutable class attributes:

    * ``PATCH_FUNCTIONS`` list to allow for "patching" when first creating
      the Fortran compiler ``F90_COMPILER`` that will be attached to the
      class (not the instances).
    * ``CUSTOM_STATIC_LIB`` callable that takes a list of object files and
      uses them to create a static / shared library. If not provided, then
      :meth:`_default_static_lib` will be used.
    * ``USE_SHARED_LIBRARY`` flag indicating if extensions will contain built
      object files or if they will refer to a shared library.
    * ``CLEANUP`` optional callable that cleans up at the end of :meth:`run`.
    """

    # Will be set at runtime, not import time.
    F90_COMPILER = None
    PATCH_FUNCTIONS = []
    CUSTOM_STATIC_LIB = None
    USE_SHARED_LIBRARY = False
    CLEANUP = None

    def __init__(self, *args, **kwargs):
        setuptools.command.build_ext.build_ext.__init__(self, *args, **kwargs)
        self.journal_file = os.environ.get(JOURNAL_ENV)
        self.commands = []

    @classmethod
    def set_f90_compiler(cls):
        import numpy.distutils.core
        import numpy.distutils.fcompiler

        if cls.F90_COMPILER is not None:
            return

        c_compiler = distutils.ccompiler.new_compiler()
        if c_compiler is None:
            return

        if c_compiler.compiler_type == "msvc":
            c_compiler.initialize()
        f90_compiler = numpy.distutils.fcompiler.new_fcompiler(
            requiref90=True, c_compiler=c_compiler
        )
        if f90_compiler is None:
            return

        dist = numpy.distutils.core.get_distribution(always=True)
        f90_compiler.customize(dist)
        # Patch up ``f90_compiler`` with any OS-specific patches.
        for patch_fn in cls.PATCH_FUNCTIONS:
            patch_fn(f90_compiler)
        cls.F90_COMPILER = f90_compiler

    @classmethod
    def has_f90_compiler(cls):
        cls.set_f90_compiler()
        return cls.F90_COMPILER is not None

    @classmethod
    def get_library_dirs(cls):
        cls.set_f90_compiler()
        if cls.USE_SHARED_LIBRARY:
            # NOTE: This assumes that the `libbezier` shared library will
            #       contain all libraries needed (e.g. there is no
            #       dependendence on ``libgfortran`` or similar). It's expected
            #       that ``library_dirs`` will be updated at run-time to have
            #       temporary build directories added.
            libraries = ["bezier"]
            library_dirs = []
            return libraries, library_dirs

        else:
            return cls.F90_COMPILER.libraries, cls.F90_COMPILER.library_dirs

    def start_journaling(self):
        """Capture calls to the system by compilers.

        See: https://github.com/numpy/numpy/blob/v1.15.2/\
        numpy/distutils/ccompiler.py#L154

        Intercepts all calls to ``CCompiler.spawn`` and keeps the
        arguments around to be stored in the local ``commands``
        instance attribute.
        """
        import numpy.distutils.ccompiler

        if self.journal_file is None:
            return

        def journaled_spawn(patched_self, cmd, display=None):
            self.commands.append(cmd)
            return numpy.distutils.ccompiler.CCompiler_spawn(
                patched_self, cmd, display=None
            )

        numpy.distutils.ccompiler.replace_method(
            distutils.ccompiler.CCompiler, "spawn", journaled_spawn
        )

    @staticmethod
    def _command_to_text(command):
        # NOTE: This assumes, but doesn't check that the command has 3
        #       or more arguments.
        first_line = "$ {} \\"
        middle_line = ">   {} \\"
        last_line = ">   {}"
        parts = [first_line.format(command[0])]
        for argument in command[1:-1]:
            parts.append(middle_line.format(argument))
        parts.append(last_line.format(command[-1]))
        return "\n".join(parts)

    def _commands_to_text(self):
        separator = "-" * 40
        parts = [separator]
        for command in self.commands:
            command_text = self._command_to_text(command)
            parts.extend([command_text, separator])
        parts.append("")  # Trailing newline in file.
        return "\n".join(parts)

    def save_journal(self):
        """Save journaled commands to file.

        If there is no active journal, does nothing.

        If saving the commands to a file fails, a message will be printed to
        STDERR but the failure will be swallowed so that the extension can
        be built successfully.
        """
        if self.journal_file is None:
            return

        try:
            as_text = self._commands_to_text()
            with open(self.journal_file, "w") as file_obj:
                file_obj.write(as_text)
        except Exception as exc:
            msg = BAD_JOURNAL.format(exc)
            print(msg, file=sys.stderr)

    def _default_static_lib(self, obj_files):
        """Create a static library (i.e. a ``.a`` / ``.lib`` file).

        Args:
            obj_files (List[str]): List of paths of compiled object files.
        """
        c_compiler = self.F90_COMPILER.c_compiler
        static_lib_dir = os.path.join(self.build_lib, "bezier", "lib")
        if not os.path.exists(static_lib_dir):
            os.makedirs(static_lib_dir)
        c_compiler.create_static_lib(
            obj_files, "bezier", output_dir=static_lib_dir
        )
        # NOTE: We must "modify" the paths for the ``extra_objects`` in
        #       each extension since they were compiled with
        #       ``output_dir=self.build_temp``.
        for extension in self.extensions:
            extension.extra_objects[:] = [
                os.path.join(self.build_temp, rel_path)
                for rel_path in extension.extra_objects
            ]

    def compile_fortran_obj_files(self):
        source_files_quadpack = [
            QUADPACK_SOURCE_FILENAME.format(mod_name)
            for mod_name in QUADPACK_MODULES
        ]
        source_files_bezier = [
            FORTRAN_SOURCE_FILENAME.format(mod_name)
            for mod_name in FORTRAN_MODULES
        ]
        source_files = source_files_quadpack + source_files_bezier
        obj_files = self.F90_COMPILER.compile(
            source_files,
            output_dir=self.build_temp,
            macros=[],
            include_dirs=[],
            debug=None,
            extra_postargs=["-J", self.build_temp],
            depends=[],
        )
        if self.CUSTOM_STATIC_LIB is None:
            self._default_static_lib(obj_files)
        else:
            self.CUSTOM_STATIC_LIB(obj_files)

    def _default_cleanup(self):
        """Default cleanup after :meth:`run`.

        For in-place builds, moves the built shared library into the source
        directory.
        """
        if not self.inplace:
            return

        shutil.move(
            os.path.join(self.build_lib, "bezier", "lib"),
            os.path.join("src", "python", "bezier"),
        )

    def cleanup(self):
        """Cleanup after :meth:`run`."""
        if self.CLEANUP is None:
            self._default_cleanup()
        else:
            self.CLEANUP()

    def copy_include_dir(self):
        """Copy ``include/`` directory into Python build directory.

        This is so the Python library will include the headers.
        """
        shutil.copytree(
            FORTRAN_INCLUDE_DIR,
            os.path.join(self.build_lib, "bezier", "include"),
        )

    def run(self):
        self.set_f90_compiler()
        self.start_journaling()
        self.copy_include_dir()
        self.compile_fortran_obj_files()
        result = setuptools.command.build_ext.build_ext.run(self)
        self.save_journal()
        self.cleanup()
        return result



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper methods for B |eacute| zier curves.

The functions provided by this module have a Cython speedup with the
exact same interface which calls out to a Fortran implementation. The speedup
will be used if the extension can be built.
"""

from bezier import _py_curve_helpers

try:
    from bezier import _speedup
except ImportError:  # pragma: NO COVER
    _speedup = None


# pylint: disable=invalid-name
if _speedup is None:  # pragma: NO COVER
    subdivide_nodes = _py_curve_helpers.subdivide_nodes
    evaluate_multi = _py_curve_helpers.evaluate_multi
    evaluate_multi_barycentric = _py_curve_helpers.evaluate_multi_barycentric
    compute_length = _py_curve_helpers.compute_length
    elevate_nodes = _py_curve_helpers.elevate_nodes
    specialize_curve = _py_curve_helpers.specialize_curve
    evaluate_hodograph = _py_curve_helpers.evaluate_hodograph
    get_curvature = _py_curve_helpers.get_curvature
    newton_refine = _py_curve_helpers.newton_refine
    locate_point = _py_curve_helpers.locate_point
    reduce_pseudo_inverse = _py_curve_helpers.reduce_pseudo_inverse
    full_reduce = _py_curve_helpers.full_reduce
else:
    subdivide_nodes = _speedup.subdivide_nodes_curve
    evaluate_multi = _speedup.evaluate_multi
    evaluate_multi_barycentric = _speedup.evaluate_multi_barycentric
    compute_length = _speedup.compute_length
    elevate_nodes = _speedup.elevate_nodes
    specialize_curve = _speedup.specialize_curve
    evaluate_hodograph = _speedup.evaluate_hodograph
    get_curvature = _speedup.get_curvature
    newton_refine = _speedup.newton_refine_curve
    locate_point = _speedup.locate_point_curve
    reduce_pseudo_inverse = _speedup.reduce_pseudo_inverse
    full_reduce = _speedup.full_reduce
# pylint: enable=invalid-name



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Cython wrapper for ``curve_intersection.f90``."""


from libcpp cimport bool as bool_t

from bezier._status cimport Status


cdef extern from "bezier/curve_intersection.h":
    cpdef enum BoxIntersectionType:
        INTERSECTION = 0
        TANGENT = 1
        DISJOINT = 2

    void newton_refine_curve_intersect "BEZ_newton_refine_curve_intersect" (
        const double* s, const int* num_nodes1,
        const double* nodes1, const double* t, const int* num_nodes2,
        const double* nodes2, double* new_s, double* new_t, Status* status)
    void bbox_intersect "BEZ_bbox_intersect" (
        const int* num_nodes1, const double* nodes1,
        const int* num_nodes2, const double* nodes2, BoxIntersectionType* enum_)
    void curve_intersections "BEZ_curve_intersections" (
        const int* num_nodes_first,
        const double* nodes_first, const int* num_nodes_second,
        const double* nodes_second, const int* intersections_size,
        double* intersections, const int* num_intersections, bool_t* coincident,
        Status* status)
    void free_curve_intersections_workspace "BEZ_free_curve_intersections_workspace" ()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Cython wrapper for ``curve.f90``."""


from libcpp cimport bool as bool_t


cdef extern from "bezier/curve.h":
    void evaluate_curve_barycentric "BEZ_evaluate_curve_barycentric" (
        const int* num_nodes, const int* dimension,
        const double* nodes, const int* num_vals, const double* lambda1,
        const double* lambda2, double* evaluated)
    void evaluate_multi "BEZ_evaluate_multi" (
        const int* num_nodes, const int* dimension,
        const double* nodes, const int* num_vals, const double* s_vals,
        double* evaluated)
    void specialize_curve "BEZ_specialize_curve" (
        const int* num_nodes, const int* dimension,
        const double* nodes, const double* start, const double* end,
        double* new_nodes)
    void evaluate_hodograph "BEZ_evaluate_hodograph" (
        double* s, const int* num_nodes,
        const int* dimension, const double* nodes, double* hodograph)
    void subdivide_nodes_curve "BEZ_subdivide_nodes_curve" (
        const int* num_nodes, const int* dimension,
        const double* nodes, double* left_nodes, double* right_nodes)
    void newton_refine_curve "BEZ_newton_refine_curve" (
        const int* num_nodes, const int* dimension,
        const double* nodes, const double* point, const double* s,
        double* updated_s)
    void locate_point_curve "BEZ_locate_point_curve" (
        const int* num_nodes, const int* dimension,
        const double* nodes, const double* point, double* s_approx)
    void elevate_nodes_curve "BEZ_elevate_nodes_curve" (
        const int* num_nodes, const int* dimension,
        const double* nodes, double* elevated)
    void get_curvature "BEZ_get_curvature" (
        const int* num_nodes, const double* nodes,
        const double* tangent_vec, const double* s, double* curvature)
    void reduce_pseudo_inverse "BEZ_reduce_pseudo_inverse" (
        const int* num_nodes, const int* dimension,
        const double* nodes, double* reduced, bool_t* not_implemented)
    void full_reduce "BEZ_full_reduce" (
        const int* num_nodes, const int* dimension,
        const double* nodes, int* num_reduced_nodes, double* reduced,
        bool_t* not_implemented)
    void compute_length "BEZ_compute_length" (
        const int* num_nodes, const int* dimension,
        const double* nodes, double* length, int* error_val)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Helper for B |eacute| zier Curves, Triangles, and Higher Order Objects.

Intended to perform basic operations on B |eacute| zier objects such
as intersections, length / area / etc. computations, subdivision,
implicitization and other relevant information.

Plotting utilities are also provided.

.. |eacute| unicode:: U+000E9 .. LATIN SMALL LETTER E WITH ACUTE
   :trim:

.. autoclass:: UnsupportedDegree
   :members:

.. autoclass:: Surface
"""

import os

import pkg_resources

# NOTE: ``__config__`` **must** be the first import because it (may)
#       modify the search path used to locate shared libraries.
from bezier import __config__
from bezier._legacy import Surface
from bezier._py_helpers import UnsupportedDegree
from bezier.curve import Curve
from bezier.curved_polygon import CurvedPolygon
from bezier.triangle import Triangle

try:
    import bezier._speedup  # noqa: F401

    _HAS_SPEEDUP = True
except ImportError as exc:  # pragma: NO COVER
    __config__.handle_import_error(exc, "_speedup")
    _HAS_SPEEDUP = False
# NOTE: The ``__version__`` and ``__author__`` are hard-coded here, rather
#       than using ``pkg_resources.get_distribution("bezier").version``
#       and related. This is **entirely** to accomodate builds where
#       ``bezier`` is imported from source (and not installed).
__author__ = "Danny Hermes"
__version__ = "2020.1.14"
"""str: The current version of :mod:`bezier`."""
__all__ = [
    "__author__",
    "__version__",
    "Curve",
    "CurvedPolygon",
    "get_dll",
    "get_include",
    "get_lib",
    "Surface",
    "Triangle",
    "UnsupportedDegree",
]


def get_include():
    """Get the directory with ``.h`` header files.

    Extension modules (and Cython modules) that need to compile against
    ``libbezier`` should use this function to locate the appropriate
    include directory.

    For more information, see :doc:`../binary-extension`.

    Returns:
        str: ``include`` directory that contains header files for the
        ``libbezier`` Fortran library.
    """
    return pkg_resources.resource_filename("bezier", "include")


def get_lib():
    """Get the directory with ``.a`` / ``.lib`` static libraries.

    Extension modules (and Cython modules) that need to compile against
    ``libbezier`` should use this function to locate the appropriate lib
    directory.

    For more information, see :doc:`../binary-extension`.

    Returns:
        str: ``lib`` directory that contains static libraries for the
        ``libbezier`` Fortran library.
    """
    return pkg_resources.resource_filename("bezier", "lib")


def get_dll():
    """Get the directory with the Windows shared library.

    Extension modules (and Cython modules) that need to compile against
    ``libbezier`` should use this function to locate the appropriate
    Windows shared library or libraries (DLLs).

    For more information, see :doc:`../binary-extension`.

    Returns:
        str: ``extra-dll`` directory that contains the Windows shared library
        for the ``libbezier`` Fortran library.

    Raises:
        OSError: If this function is used anywhere other than Windows.
    """
    if os.name == "nt":
        return pkg_resources.resource_filename("bezier", "extra-dll")

    else:
        raise OSError("This function should only be used on Windows.")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pure Python implementations of helper methods for B |eacute| zier curves.

.. |eacute| unicode:: U+000E9 .. LATIN SMALL LETTER E WITH ACUTE
   :trim:
"""

import functools

import numpy as np

try:
    import scipy.integrate as _scipy_int
except ImportError:  # pragma: NO COVER
    _scipy_int = None

from bezier import _py_helpers


_MAX_LOCATE_SUBDIVISIONS = 20
_LOCATE_STD_CAP = 0.5 ** 20
_FLOAT64 = np.float64  # pylint: disable=no-member
_REDUCE_THRESHOLD = 0.5 ** 26  # sqrt(machine precision)
# Projections onto the space of degree-elevated nodes.
# If v --> vE is the (right) elevation map, then P = E^T (E E^T)^{-1} E
# is the (right) projection.
_PROJECTION0 = np.asfortranarray([[0.5, 0.5], [0.5, 0.5]])
_PROJ_DENOM0 = 1.0
_PROJECTION1 = np.asfortranarray(
    [[2.5, 1.0, -0.5], [1.0, 1.0, 1.0], [-0.5, 1.0, 2.5]]
)
_PROJ_DENOM1 = 3.0
_PROJECTION2 = np.asfortranarray(
    [
        [4.75, 0.75, -0.75, 0.25],
        [0.75, 2.75, 2.25, -0.75],
        [-0.75, 2.25, 2.75, 0.75],
        [0.25, -0.75, 0.75, 4.75],
    ]
)
_PROJ_DENOM2 = 5.0
_PROJECTION3 = np.asfortranarray(
    [
        [34.5, 2.0, -3.0, 2.0, -0.5],
        [2.0, 27.0, 12.0, -8.0, 2.0],
        [-3.0, 12.0, 17.0, 12.0, -3.0],
        [2.0, -8.0, 12.0, 27.0, 2.0],
        [-0.5, 2.0, -3.0, 2.0, 34.5],
    ]
)
_PROJ_DENOM3 = 35.0
# Reductions for a set of degree-elevated nodes.
# If v --> vE is the (right) elevation map, then R = E^T (E E^T)^{-1} -- the
# (right) pseudo-inverse of E -- actually reduces a set of nodes.
_REDUCTION0 = np.asfortranarray([[0.5], [0.5]])
_REDUCTION_DENOM0 = 1.0
_REDUCTION1 = np.asfortranarray([[2.5, -0.5], [1.0, 1.0], [-0.5, 2.5]])
_REDUCTION_DENOM1 = 3.0
_REDUCTION2 = np.asfortranarray(
    [
        [4.75, -1.25, 0.25],
        [0.75, 3.75, -0.75],
        [-0.75, 3.75, 0.75],
        [0.25, -1.25, 4.75],
    ]
)
_REDUCTION_DENOM2 = 5.0
_REDUCTION3 = np.asfortranarray(
    [
        [103.5, -26.5, 8.5, -1.5],
        [6.0, 106.0, -34.0, 6.0],
        [-9.0, 51.0, 51.0, -9.0],
        [6.0, -34.0, 106.0, 6.0],
        [-1.5, 8.5, -26.5, 103.5],
    ]
)
_REDUCTION_DENOM3 = 105.0
_LINEAR_SUBDIVIDE_LEFT = np.asfortranarray([[1.0, 0.5], [0.0, 0.5]])
_LINEAR_SUBDIVIDE_RIGHT = np.asfortranarray([[0.5, 0.0], [0.5, 1.0]])
_QUADRATIC_SUBDIVIDE_LEFT = np.asfortranarray(
    [[1.0, 0.5, 0.25], [0.0, 0.5, 0.5], [0.0, 0.0, 0.25]]
)
_QUADRATIC_SUBDIVIDE_RIGHT = np.asfortranarray(
    [[0.25, 0.0, 0.0], [0.5, 0.5, 0.0], [0.25, 0.5, 1.0]]
)
_CUBIC_SUBDIVIDE_LEFT = np.asfortranarray(
    [
        [1.0, 0.5, 0.25, 0.125],
        [0.0, 0.5, 0.5, 0.375],
        [0.0, 0.0, 0.25, 0.375],
        [0.0, 0.0, 0.0, 0.125],
    ]
)
_CUBIC_SUBDIVIDE_RIGHT = np.asfortranarray(
    [
        [0.125, 0.0, 0.0, 0.0],
        [0.375, 0.25, 0.0, 0.0],
        [0.375, 0.5, 0.5, 0.0],
        [0.125, 0.25, 0.5, 1.0],
    ]
)


def make_subdivision_matrices(degree):
    """Make the matrix used to subdivide a curve.

    .. note::

        This is a helper for :func:`subdivide_nodes`. It does not have a
        Fortran speedup because it is **only** used by a function which has
        a Fortran speedup.

    Args:
        degree (int): The degree of the curve.

    Returns:
        Tuple[numpy.ndarray, numpy.ndarray]: The matrices used to convert
           the nodes into left and right nodes, respectively.
    """
    left = np.zeros((degree + 1, degree + 1), order="F")
    right = np.zeros((degree + 1, degree + 1), order="F")
    left[0, 0] = 1.0
    right[-1, -1] = 1.0
    for col in range(1, degree + 1):
        half_prev = 0.5 * left[:col, col - 1]
        left[:col, col] = half_prev
        left[1 : col + 1, col] += half_prev  # noqa: E203
        # Populate the complement col (in right) as well.
        complement = degree - col
        # NOTE: We "should" reverse the results when using
        #       the complement, but they are symmetric so
        #       that would be a waste.
        right[-(col + 1) :, complement] = left[: col + 1, col]  # noqa: E203
    return left, right


def subdivide_nodes(nodes):
    """Subdivide a curve into two sub-curves.

    Does so by taking the unit interval (i.e. the domain of the curve) and
    splitting it into two sub-intervals by splitting down the middle.

    .. note::

       There is also a Fortran implementation of this function, which
       will be used if it can be built.

    Args:
        nodes (numpy.ndarray): The nodes defining a B |eacute| zier curve.

    Returns:
        Tuple[numpy.ndarray, numpy.ndarray]: The nodes for the two sub-curves.
    """
    _, num_nodes = np.shape(nodes)
    if num_nodes == 2:
        left_nodes = _py_helpers.matrix_product(nodes, _LINEAR_SUBDIVIDE_LEFT)
        right_nodes = _py_helpers.matrix_product(
            nodes, _LINEAR_SUBDIVIDE_RIGHT
        )
    elif num_nodes == 3:
        left_nodes = _py_helpers.matrix_product(
            nodes, _QUADRATIC_SUBDIVIDE_LEFT
        )
        right_nodes = _py_helpers.matrix_product(
            nodes, _QUADRATIC_SUBDIVIDE_RIGHT
        )
    elif num_nodes == 4:
        left_nodes = _py_helpers.matrix_product(nodes, _CUBIC_SUBDIVIDE_LEFT)
        right_nodes = _py_helpers.matrix_product(nodes, _CUBIC_SUBDIVIDE_RIGHT)
    else:
        left_mat, right_mat = make_subdivision_matrices(num_nodes - 1)
        left_nodes = _py_helpers.matrix_product(nodes, left_mat)
        right_nodes = _py_helpers.matrix_product(nodes, right_mat)
    return left_nodes, right_nodes


def evaluate_multi(nodes, s_vals):
    r"""Computes multiple points along a curve.

    Does so via a modified Horner's method for each value in ``s_vals``
    rather than using the de Casteljau algorithm.

    .. note::

       There is also a Fortran implementation of this function, which
       will be used if it can be built.

    Args:
        nodes (numpy.ndarray): The nodes defining a curve.
        s_vals (numpy.ndarray): Parameters along the curve (as a
            1D array).

    Returns:
        numpy.ndarray: The evaluated points on the curve as a two dimensional
        NumPy array, with the columns corresponding to each ``s``
        value and the rows to the dimension.
    """
    one_less = 1.0 - s_vals
    return evaluate_multi_barycentric(nodes, one_less, s_vals)


def evaluate_multi_barycentric(nodes, lambda1, lambda2):
    r"""Evaluates a B |eacute| zier type-function.

    Of the form

    .. math::

       B(\lambda_1, \lambda_2) = \sum_j \binom{n}{j}
           \lambda_1^{n - j} \lambda_2^j \cdot v_j

    for some set of vectors :math:`v_j` given by ``nodes``.

    Does so via a modified Horner's method for each pair of values
    in ``lambda1`` and ``lambda2``, rather than using the
    de Casteljau algorithm.

    .. note::

       There is also a Fortran implementation of this function, which
       will be used if it can be built.

    Args:
        nodes (numpy.ndarray): The nodes defining a curve.
        lambda1 (numpy.ndarray): Parameters along the curve (as a
            1D array).
        lambda2 (numpy.ndarray): Parameters along the curve (as a
            1D array). Typically we have ``lambda1 + lambda2 == 1``.

    Returns:
        numpy.ndarray: The evaluated points as a two dimensional
        NumPy array, with the columns corresponding to each pair of parameter
        values and the rows to the dimension.
    """
    # NOTE: We assume but don't check that lambda2 has the same shape.
    (num_vals,) = lambda1.shape
    dimension, num_nodes = nodes.shape
    degree = num_nodes - 1
    # Resize as row vectors for broadcast multiplying with
    # columns of ``nodes``.
    lambda1 = lambda1[np.newaxis, :]
    lambda2 = lambda2[np.newaxis, :]
    result = np.zeros((dimension, num_vals), order="F")
    result += lambda1 * nodes[:, [0]]
    binom_val = 1.0
    lambda2_pow = np.ones((1, num_vals), order="F")
    for index in range(1, degree):
        lambda2_pow *= lambda2
        binom_val = (binom_val * (degree - index + 1)) / index
        result += binom_val * lambda2_pow * nodes[:, [index]]
        result *= lambda1
    result += lambda2 * lambda2_pow * nodes[:, [degree]]
    return result


def vec_size(nodes, s_val):
    r"""Compute :math:`\|B(s)\|_2`.

    .. note::

        This is a helper for :func:`compute_length` and does not have
        a Fortran speedup.

    Intended to be used with ``functools.partial`` to fill in the
    value of ``nodes`` and create a callable that only accepts ``s_val``.

    Args:
        nodes (numpy.ndarray): The nodes defining a curve.
        s_val (float): Parameter to compute :math:`B(s)`.

    Returns:
        float: The norm of :math:`B(s)`.
    """
    result_vec = evaluate_multi(nodes, np.asfortranarray([s_val]))
    # NOTE: We convert to 1D to make sure NumPy uses vector norm.
    return np.linalg.norm(result_vec[:, 0], ord=2)


def compute_length(nodes):
    r"""Approximately compute the length of a curve.

    .. _QUADPACK: https://en.wikipedia.org/wiki/QUADPACK

    If ``degree`` is :math:`n`, then the Hodograph curve
    :math:`B'(s)` is degree :math:`d = n - 1`. Using this curve, we
    approximate the integral:

    .. math::

       \int_{B\left(\left[0, 1\right]\right)} 1 \, d\mathbf{x} =
       \int_0^1 \left\lVert B'(s) \right\rVert_2 \, ds

    using `QUADPACK`_ (via SciPy).

    .. note::

       There is also a Fortran implementation of this function, which
       will be used if it can be built.

    Args:
        nodes (numpy.ndarray): The nodes defining a curve.

    Returns:
        float: The length of the curve.

    Raises:
        ValueError: If ``nodes`` has zero columns.
        OSError: If SciPy is not installed.
    """
    _, num_nodes = np.shape(nodes)
    # NOTE: We somewhat replicate code in ``evaluate_hodograph()``
    #       here. This is so we don't re-compute the nodes for the first
    #       derivative every time it is evaluated.
    first_deriv = (num_nodes - 1) * (nodes[:, 1:] - nodes[:, :-1])
    if num_nodes == 0:
        raise ValueError("Curve should have at least one node.")

    if num_nodes == 1:
        return 0.0

    if num_nodes == 2:
        # NOTE: We convert to 1D to make sure NumPy uses vector norm.
        return np.linalg.norm(first_deriv[:, 0], ord=2)

    if _scipy_int is None:
        raise OSError("This function requires SciPy for quadrature.")

    size_func = functools.partial(vec_size, first_deriv)
    length, _ = _scipy_int.quad(size_func, 0.0, 1.0)
    return length


def elevate_nodes(nodes):
    r"""Degree-elevate a B |eacute| zier curves.

    Does this by converting the current nodes :math:`v_0, \ldots, v_n`
    to new nodes :math:`w_0, \ldots, w_{n + 1}` where

    .. math::

       \begin{align*}
       w_0 &= v_0 \\
       w_j &= \frac{j}{n + 1} v_{j - 1} + \frac{n + 1 - j}{n + 1} v_j \\
       w_{n + 1} &= v_n
       \end{align*}

    .. note::

       There is also a Fortran implementation of this function, which
       will be used if it can be built.

    Args:
        nodes (numpy.ndarray): The nodes defining a curve.

    Returns:
        numpy.ndarray: The nodes of the degree-elevated curve.
    """
    dimension, num_nodes = np.shape(nodes)
    new_nodes = np.empty((dimension, num_nodes + 1), order="F")
    multipliers = np.arange(1, num_nodes, dtype=_FLOAT64)[np.newaxis, :]
    denominator = float(num_nodes)
    new_nodes[:, 1:-1] = (
        multipliers * nodes[:, :-1]
        + (denominator - multipliers) * nodes[:, 1:]
    )
    # Hold off on division until the end, to (attempt to) avoid round-off.
    new_nodes /= denominator
    # After setting the internal nodes (which require division), set the
    # boundary nodes.
    new_nodes[:, 0] = nodes[:, 0]
    new_nodes[:, -1] = nodes[:, -1]
    return new_nodes


def de_casteljau_one_round(nodes, lambda1, lambda2):
    """Perform one round of de Casteljau's algorithm.

    .. note::

        This is a helper for :func:`specialize_curve`. It does not have a
        Fortran speedup because it is **only** used by a function which has
        a Fortran speedup.

    The weights are assumed to sum to one.

    Args:
        nodes (numpy.ndarray): Control points for a curve.
        lambda1 (float): First barycentric weight on interval.
        lambda2 (float): Second barycentric weight on interval.

    Returns:
        numpy.ndarray: The nodes for a "blended" curve one degree
        lower.
    """
    return np.asfortranarray(lambda1 * nodes[:, :-1] + lambda2 * nodes[:, 1:])


def specialize_curve(nodes, start, end):
    """Specialize a curve to a re-parameterization

    .. note::

       This assumes the curve is degree 1 or greater but doesn't check.

    .. note::

       There is also a Fortran implementation of this function, which
       will be used if it can be built.

    Args:
        nodes (numpy.ndarray): Control points for a curve.
        start (float): The start point of the interval we are specializing to.
        end (float): The end point of the interval we are specializing to.

    Returns:
        numpy.ndarray: The control points for the specialized curve.
    """
    # NOTE: There is no corresponding "enable", but the disable only applies
    #       in this lexical scope.
    # pylint: disable=too-many-locals
    _, num_nodes = np.shape(nodes)
    # Uses start-->0, end-->1 to represent the specialization used.
    weights = ((1.0 - start, start), (1.0 - end, end))
    partial_vals = {
        (0,): de_casteljau_one_round(nodes, *weights[0]),
        (1,): de_casteljau_one_round(nodes, *weights[1]),
    }
    for _ in range(num_nodes - 2, 0, -1):
        new_partial = {}
        for key, sub_nodes in partial_vals.items():
            # Our keys are ascending so we increment from the last value.
            for next_id in range(key[-1], 1 + 1):
                new_key = key + (next_id,)
                new_partial[new_key] = de_casteljau_one_round(
                    sub_nodes, *weights[next_id]
                )
        partial_vals = new_partial
    result = np.empty(nodes.shape, order="F")
    for index in range(num_nodes):
        key = (0,) * (num_nodes - index - 1) + (1,) * index
        result[:, [index]] = partial_vals[key]
    return result


def evaluate_hodograph(s, nodes):
    r"""Evaluate the Hodograph curve at a point :math:`s`.

    The Hodograph (first derivative) of a B |eacute| zier curve
    degree :math:`d = n - 1` and is given by

    .. math::

       B'(s) = n \sum_{j = 0}^{d} \binom{d}{j} s^j
       (1 - s)^{d - j} \cdot \Delta v_j

    where each forward difference is given by
    :math:`\Delta v_j = v_{j + 1} - v_j`.

    .. note::

       There is also a Fortran implementation of this function, which
       will be used if it can be built.

    Args:
        nodes (numpy.ndarray): The nodes of a curve.
        s (float): A parameter along the curve at which the Hodograph
            is to be evaluated.

    Returns:
        numpy.ndarray: The point on the Hodograph curve (as a two
        dimensional NumPy array with a single row).
    """
    _, num_nodes = np.shape(nodes)
    first_deriv = nodes[:, 1:] - nodes[:, :-1]
    return (num_nodes - 1) * evaluate_multi(
        first_deriv, np.asfortranarray([s])
    )


def get_curvature(nodes, tangent_vec, s):
    r"""Compute the signed curvature of a curve at :math:`s`.

    Computed via

    .. math::

       \frac{B'(s) \times B''(s)}{\left\lVert B'(s) \right\rVert_2^3}

    .. image:: ../images/get_curvature.png
       :align: center

    .. testsetup:: get-curvature

       import numpy as np
       import bezier
       from bezier._py_curve_helpers import evaluate_hodograph
       from bezier._py_curve_helpers import get_curvature

    .. doctest:: get-curvature
       :options: +NORMALIZE_WHITESPACE

       >>> nodes = np.asfortranarray([
       ...     [1.0, 0.75,  0.5, 0.25, 0.0],
       ...     [0.0, 2.0 , -2.0, 2.0 , 0.0],
       ... ])
       >>> s = 0.5
       >>> tangent_vec = evaluate_hodograph(s, nodes)
       >>> tangent_vec
       array([[-1.],
              [ 0.]])
       >>> curvature = get_curvature(nodes, tangent_vec, s)
       >>> curvature
       -12.0

    .. testcleanup:: get-curvature

       import make_images
       make_images.get_curvature(nodes, s, tangent_vec, curvature)

    .. note::

       There is also a Fortran implementation of this function, which
       will be used if it can be built.

    Args:
        nodes (numpy.ndarray): The nodes of a curve.
        tangent_vec (numpy.ndarray): The already computed value of
            :math:`B'(s)`
        s (float): The parameter value along the curve.

    Returns:
        float: The signed curvature.
    """
    _, num_nodes = np.shape(nodes)
    if num_nodes == 2:  # Lines have no curvature.
        return 0.0

    # NOTE: We somewhat replicate code in ``evaluate_hodograph()`` here.
    first_deriv = nodes[:, 1:] - nodes[:, :-1]
    second_deriv = first_deriv[:, 1:] - first_deriv[:, :-1]
    concavity = (
        (num_nodes - 1)
        * (num_nodes - 2)
        * evaluate_multi(second_deriv, np.asfortranarray([s]))
    )
    curvature = _py_helpers.cross_product(
        tangent_vec.ravel(order="F"), concavity.ravel(order="F")
    )
    # NOTE: We convert to 1D to make sure NumPy uses vector norm.
    curvature /= np.linalg.norm(tangent_vec[:, 0], ord=2) ** 3
    return curvature


def newton_refine(nodes, point, s):
    r"""Refine a solution to :math:`B(s) = p` using Newton's method.

    Computes updates via

    .. math::

       \mathbf{0} \approx
           \left(B\left(s_{\ast}\right) - p\right) +
           B'\left(s_{\ast}\right) \Delta s

    For example, consider the curve

    .. math::

       B(s) =
           \left[\begin{array}{c} 0 \\ 0 \end{array}\right] (1 - s)^2 +
           \left[\begin{array}{c} 1 \\ 2 \end{array}\right]
               2 (1 - s) s +
           \left[\begin{array}{c} 3 \\ 1 \end{array}\right] s^2

    and the point :math:`B\left(\frac{1}{4}\right) =
    \frac{1}{16} \left[\begin{array}{c} 9 \\ 13 \end{array}\right]`.

    Starting from the **wrong** point :math:`s = \frac{3}{4}`, we have

    .. math::

       \begin{align*}
       p - B\left(\frac{1}{2}\right) &= -\frac{1}{2}
           \left[\begin{array}{c} 3 \\ 1 \end{array}\right] \\
       B'\left(\frac{1}{2}\right) &= \frac{1}{2}
           \left[\begin{array}{c} 7 \\ -1 \end{array}\right] \\
       \Longrightarrow \frac{1}{4} \left[\begin{array}{c c}
           7 & -1 \end{array}\right] \left[\begin{array}{c}
           7 \\ -1 \end{array}\right] \Delta s &= -\frac{1}{4}
           \left[\begin{array}{c c} 7 & -1 \end{array}\right]
           \left[\begin{array}{c} 3 \\ 1 \end{array}\right] \\
       \Longrightarrow \Delta s &= -\frac{2}{5}
       \end{align*}

    .. image:: ../images/newton_refine_curve.png
       :align: center

    .. testsetup:: newton-refine-curve, newton-refine-curve-cusp

       import numpy as np
       import bezier
       from bezier._py_curve_helpers import newton_refine

    .. doctest:: newton-refine-curve
       :options: +NORMALIZE_WHITESPACE

       >>> nodes = np.asfortranarray([
       ...     [0.0, 1.0, 3.0],
       ...     [0.0, 2.0, 1.0],
       ... ])
       >>> curve = bezier.Curve(nodes, degree=2)
       >>> point = curve.evaluate(0.25)
       >>> point
       array([[0.5625],
              [0.8125]])
       >>> s = 0.75
       >>> new_s = newton_refine(nodes, point, s)
       >>> 5 * (new_s - s)
       -2.0

    .. testcleanup:: newton-refine-curve

       import make_images
       make_images.newton_refine_curve(curve, point, s, new_s)

    On curves that are not "valid" (i.e. :math:`B(s)` is not
    injective with non-zero gradient), Newton's method may
    break down and converge linearly:

    .. image:: ../images/newton_refine_curve_cusp.png
       :align: center

    .. doctest:: newton-refine-curve-cusp
       :options: +NORMALIZE_WHITESPACE

       >>> nodes = np.asfortranarray([
       ...     [ 6.0, -2.0, -2.0, 6.0],
       ...     [-3.0,  3.0, -3.0, 3.0],
       ... ])
       >>> curve = bezier.Curve(nodes, degree=3)
       >>> expected = 0.5
       >>> point = curve.evaluate(expected)
       >>> point
       array([[0.],
              [0.]])
       >>> s_vals = [0.625, None, None, None, None, None]
       >>> np.log2(abs(expected - s_vals[0]))
       -3.0
       >>> s_vals[1] = newton_refine(nodes, point, s_vals[0])
       >>> np.log2(abs(expected - s_vals[1]))
       -3.983...
       >>> s_vals[2] = newton_refine(nodes, point, s_vals[1])
       >>> np.log2(abs(expected - s_vals[2]))
       -4.979...
       >>> s_vals[3] = newton_refine(nodes, point, s_vals[2])
       >>> np.log2(abs(expected - s_vals[3]))
       -5.978...
       >>> s_vals[4] = newton_refine(nodes, point, s_vals[3])
       >>> np.log2(abs(expected - s_vals[4]))
       -6.978...
       >>> s_vals[5] = newton_refine(nodes, point, s_vals[4])
       >>> np.log2(abs(expected - s_vals[5]))
       -7.978...

    .. testcleanup:: newton-refine-curve-cusp

       import make_images
       make_images.newton_refine_curve_cusp(curve, s_vals)

    Due to round-off, the Newton process terminates with an error that is not
    close to machine precision :math:`\varepsilon` when :math:`\Delta s = 0`.

    .. testsetup:: newton-refine-curve-cusp-continued

       import numpy as np
       import bezier
       from bezier._py_curve_helpers import newton_refine

       nodes = np.asfortranarray([
           [ 6.0, -2.0, -2.0, 6.0],
           [-3.0,  3.0, -3.0, 3.0],
       ])
       curve = bezier.Curve(nodes, degree=3)
       point = curve.evaluate(0.5)

    .. doctest:: newton-refine-curve-cusp-continued

       >>> s_vals = [0.625]
       >>> new_s = newton_refine(nodes, point, s_vals[-1])
       >>> while new_s not in s_vals:
       ...     s_vals.append(new_s)
       ...     new_s = newton_refine(nodes, point, s_vals[-1])
       ...
       >>> terminal_s = s_vals[-1]
       >>> terminal_s == newton_refine(nodes, point, terminal_s)
       True
       >>> 2.0**(-31) <= abs(terminal_s - 0.5) <= 2.0**(-28)
       True

    Due to round-off near the cusp, the final error resembles
    :math:`\sqrt{\varepsilon}` rather than machine precision
    as expected.

    .. note::

       There is also a Fortran implementation of this function, which
       will be used if it can be built.

    Args:
        nodes (numpy.ndarray): The nodes defining a B |eacute| zier curve.
        point (numpy.ndarray): A point on the curve.
        s (float): An "almost" solution to :math:`B(s) = p`.

    Returns:
        float: The updated value :math:`s + \Delta s`.
    """
    pt_delta = point - evaluate_multi(nodes, np.asfortranarray([s]))
    derivative = evaluate_hodograph(s, nodes)
    # Each array is 2 x 1 (i.e. a column vector), we want the vector
    # dot product.
    delta_s = np.vdot(pt_delta[:, 0], derivative[:, 0]) / np.vdot(
        derivative[:, 0], derivative[:, 0]
    )
    return s + delta_s


def locate_point(nodes, point):
    r"""Locate a point on a curve.

    Does so by recursively subdividing the curve and rejecting
    sub-curves with bounding boxes that don't contain the point.
    After the sub-curves are sufficiently small, uses Newton's
    method to zoom in on the parameter value.

    .. note::

       This assumes, but does not check, that ``point`` is ``D x 1``,
       where ``D`` is the dimension that ``curve`` is in.

    .. note::

       There is also a Fortran implementation of this function, which
       will be used if it can be built.

    Args:
        nodes (numpy.ndarray): The nodes defining a B |eacute| zier curve.
        point (numpy.ndarray): The point to locate.

    Returns:
        Optional[float]: The parameter value (:math:`s`) corresponding
        to ``point`` or :data:`None` if the point is not on the ``curve``.

    Raises:
        ValueError: If the standard deviation of the remaining start / end
            parameters among the subdivided intervals exceeds a given
            threshold (e.g. :math:`2^{-20}`).
    """
    candidates = [(0.0, 1.0, nodes)]
    for _ in range(_MAX_LOCATE_SUBDIVISIONS + 1):
        next_candidates = []
        for start, end, candidate in candidates:
            if _py_helpers.contains_nd(candidate, point.ravel(order="F")):
                midpoint = 0.5 * (start + end)
                left, right = subdivide_nodes(candidate)
                next_candidates.extend(
                    ((start, midpoint, left), (midpoint, end, right))
                )
        candidates = next_candidates
    if not candidates:
        return None

    params = [(start, end) for start, end, _ in candidates]
    if np.std(params) > _LOCATE_STD_CAP:
        raise ValueError("Parameters not close enough to one another", params)

    s_approx = np.mean(params)
    s_approx = newton_refine(nodes, point, s_approx)
    # NOTE: Since ``np.mean(params)`` must be in ``[0, 1]`` it's
    #       "safe" to push the Newton-refined value back into the unit
    #       interval.
    if s_approx < 0.0:
        return 0.0

    elif s_approx > 1.0:
        return 1.0

    else:
        return s_approx


def reduce_pseudo_inverse(nodes):
    """Performs degree-reduction for a B |eacute| zier curve.

    Does so by using the pseudo-inverse of the degree elevation
    operator (which is overdetermined).

    .. note::

       There is also a Fortran implementation of this function, which
       will be used if it can be built.

    Args:
        nodes (numpy.ndarray): The nodes in the curve.

    Returns:
        numpy.ndarray: The reduced nodes.

    Raises:
        .UnsupportedDegree: If the degree is not 1, 2, 3 or 4.
    """
    _, num_nodes = np.shape(nodes)
    if num_nodes == 2:
        reduction = _REDUCTION0
        denom = _REDUCTION_DENOM0
    elif num_nodes == 3:
        reduction = _REDUCTION1
        denom = _REDUCTION_DENOM1
    elif num_nodes == 4:
        reduction = _REDUCTION2
        denom = _REDUCTION_DENOM2
    elif num_nodes == 5:
        reduction = _REDUCTION3
        denom = _REDUCTION_DENOM3
    else:
        raise _py_helpers.UnsupportedDegree(
            num_nodes - 1, supported=(1, 2, 3, 4)
        )

    result = _py_helpers.matrix_product(nodes, reduction)
    result /= denom
    return result


def projection_error(nodes, projected):
    """Compute the error between ``nodes`` and the projected nodes.

    .. note::

        This is a helper for :func:`maybe_reduce`, which is in turn a helper
        for :func:`_full_reduce`. Hence there is no corresponding Fortran
        speedup.

    For now, just compute the relative error in the Frobenius norm. But,
    we may wish to consider the error per row / point instead.

    Args:
        nodes (numpy.ndarray): Nodes in a curve.
        projected (numpy.ndarray): The ``nodes`` projected into the
            space of degree-elevated nodes.

    Returns:
        float: The relative error.
    """
    relative_err = np.linalg.norm(nodes - projected, ord="fro")
    if relative_err != 0.0:
        relative_err /= np.linalg.norm(nodes, ord="fro")
    return relative_err


def maybe_reduce(nodes):
    r"""Reduce nodes in a curve if they are degree-elevated.

    .. note::

        This is a helper for :func:`_full_reduce`. Hence there is no
        corresponding Fortran speedup.

    We check if the nodes are degree-elevated by projecting onto the
    space of degree-elevated curves of the same degree, then comparing
    to the projection. We form the projection by taking the corresponding
    (right) elevation matrix :math:`E` (from one degree lower) and forming
    :math:`E^T \left(E E^T\right)^{-1} E`.

    Args:
        nodes (numpy.ndarray): The nodes in the curve.

    Returns:
        Tuple[bool, numpy.ndarray]: Pair of values. The first indicates
        if the ``nodes`` were reduced. The second is the resulting nodes,
        either the reduced ones or the original passed in.

    Raises:
        .UnsupportedDegree: If the curve is degree 5 or higher.
    """
    _, num_nodes = nodes.shape
    if num_nodes < 2:
        return False, nodes

    elif num_nodes == 2:
        projection = _PROJECTION0
        denom = _PROJ_DENOM0
    elif num_nodes == 3:
        projection = _PROJECTION1
        denom = _PROJ_DENOM1
    elif num_nodes == 4:
        projection = _PROJECTION2
        denom = _PROJ_DENOM2
    elif num_nodes == 5:
        projection = _PROJECTION3
        denom = _PROJ_DENOM3
    else:
        raise _py_helpers.UnsupportedDegree(
            num_nodes - 1, supported=(0, 1, 2, 3, 4)
        )

    projected = _py_helpers.matrix_product(nodes, projection) / denom
    relative_err = projection_error(nodes, projected)
    if relative_err < _REDUCE_THRESHOLD:
        return True, reduce_pseudo_inverse(nodes)

    else:
        return False, nodes


def full_reduce(nodes):
    """Apply degree reduction to ``nodes`` until it can no longer be reduced.

    .. note::

       There is also a Fortran implementation of this function, which
       will be used if it can be built.

    Args:
        nodes (numpy.ndarray): The nodes in the curve.

    Returns:
        numpy.ndarray: The fully degree-reduced nodes.
    """
    was_reduced, nodes = maybe_reduce(nodes)
    while was_reduced:
        was_reduced, nodes = maybe_reduce(nodes)
    return nodes



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper for doing symbolic algebra with data.

Includes functions for:

* Converting floating point arrays to rational SymPy matrices
* Computing B |eacute| zier curve and triangle polynomial representations
* Implicitizing parametric equations

.. |eacute| unicode:: U+000E9 .. LATIN SMALL LETTER E WITH ACUTE
   :trim:
"""

import functools

try:
    import sympy
except ImportError:  # pragma: NO COVER
    sympy = None


def require_sympy(wrapped):
    """Function decorator to require ``sympy`` to exist.

    Args:
        wrapped (Callable): A function to be wrapped.

    Returns:
        Callable: The wrapped function.
    """

    @functools.wraps(wrapped)
    def ensure_sympy(*args, **kwargs):
        if sympy is None:
            raise OSError("This function requires SymPy.")
        return wrapped(*args, **kwargs)

    return ensure_sympy


@require_sympy
def to_symbolic(nodes):
    """Convert a 2D NumPy array to a SymPy matrix of rational numbers.

    Args:
        nodes (numpy.ndarray): Nodes defining an object, a 2D NumPy array.

    Returns:
        sympy.Matrix: The nodes as a SymPy matrix of rational numbers.

    Raises:
        ValueError: If ``nodes`` is not 2D.
    """
    if nodes.ndim != 2:
        raise ValueError("Nodes must be 2-dimensional, not", nodes.ndim)

    return sympy.Matrix(
        [[sympy.Rational(value) for value in row] for row in nodes]
    )


@require_sympy
def curve_weights(degree, s):
    """Compute de Casteljau weights for a curve.

    .. note::

       This function could be optimized by introducing a cache. However, any
       code using SymPy is (for now) assumed not to be performance critical.

    Args:
        degree (int): The degree of a curve.
        s (sympy.Symbol): The symbol to be used in the weights.

    Returns:
        sympy.Matrix: The de Casteljau weights for the curve as a
        ``(degree + 1) x 1`` matrix.
    """
    return sympy.Matrix(
        [
            [sympy.binomial(degree, k) * s ** k * (1 - s) ** (degree - k)]
            for k in range(degree + 1)
        ]
    )


@require_sympy
def curve_as_polynomial(nodes, degree):
    """Convert ``nodes`` into a SymPy polynomial array :math:`B(s)`.

    Args:
        nodes (numpy.ndarray): Nodes defining a B |eacute| zier curve.
        degree (int): The degree of the curve. This is assumed to
            correctly correspond to the number of ``nodes``.

    Returns:
        Tuple[sympy.Symbol, sympy.Matrix]: Pair of
        * The symbol ``s`` used in the polynomial
        * The curve :math:`B(s)`.
    """
    nodes_sym = to_symbolic(nodes)

    s = sympy.Symbol("s")
    b_polynomial = nodes_sym * curve_weights(degree, s)
    b_polynomial.simplify()

    factored = [value.factor() for value in b_polynomial]
    return s, sympy.Matrix(factored).reshape(*b_polynomial.shape)


@require_sympy
def implicitize_2d(x_fn, y_fn, s):
    """Implicitize a 2D parametric curve.

    Args:
        x_fn (sympy.Expr): Function :math:`x(s)` in the curve.
        y_fn (sympy.Expr): Function :math:`y(s)` in the curve.
        s (sympy.Symbol): The symbol used to define ``x_fn`` and ``y_fn``.

    Returns:
        sympy.Expr: The implicitized function :math:`f(x, y)` such that the
        curve satisfies :math:`f(x(s), y(s)) = 0`.
    """
    x_sym, y_sym = sympy.symbols("x, y")
    return sympy.resultant(x_fn - x_sym, y_fn - y_sym, s).factor()


@require_sympy
def implicitize_curve(nodes, degree):
    """Implicitize a 2D parametric curve, given the nodes.

    .. note::

       This function assumes (but does not check) that the caller has verified
       ``nodes`` represents a 2D curve. If it **does not**, this function will
       error out when tuple-unpacking the polynomial.

    Args:
        nodes (numpy.ndarray): Nodes defining a B |eacute| zier curve.
        degree (int): The degree of the curve. This is assumed to
            correctly correspond to the number of ``nodes``.

    Returns:
        sympy.Expr: The implicitized function :math:`f(x, y)` such that the
        curve satisfies :math:`f(x(s), y(s)) = 0`.
    """
    s, b_polynomial = curve_as_polynomial(nodes, degree)
    x_fn, y_fn = b_polynomial
    return implicitize_2d(x_fn, y_fn, s)


@require_sympy
def triangle_weights(degree, s, t):
    """Compute de Casteljau weights for a triangle.

    .. note::

       This function could be optimized by introducing a cache. However, any
       code using SymPy is (for now) assumed not to be performance critical.

    Args:
        degree (int): The degree of a triangle.
        s (sympy.Symbol): The first symbol to be used in the weights.
        t (sympy.Symbol): The second symbol to be used in the weights.

    Returns:
        sympy.Matrix: The de Casteljau weights for the triangle as an ``N x 1``
        matrix, where ``N == (degree + 1)(degree + 2) / 2``.
    """
    lambda1 = 1 - s - t
    lambda2 = s
    lambda3 = t

    values = []
    for k in range(degree + 1):
        coeff_k = sympy.binomial(degree, k)
        # N! / i! j! k! = N! / [k! (N - k)!] (i + j)! / [i! j!]
        for j in range(degree - k + 1):
            i = degree - j - k
            coeff = coeff_k * sympy.binomial(degree - k, j)
            values.append(coeff * lambda1 ** i * lambda2 ** j * lambda3 ** k)

    return sympy.Matrix(values).reshape(len(values), 1)


@require_sympy
def triangle_as_polynomial(nodes, degree):
    """Convert ``nodes`` into a SymPy polynomial array :math:`B(s, t)`.

    Args:
        nodes (numpy.ndarray): Nodes defining a B |eacute| zier triangle.
        degree (int): The degree of the triangle. This is assumed to
            correctly correspond to the number of ``nodes``.

    Returns:
        Tuple[sympy.Symbol, sympy.Symbol, sympy.Matrix]: Triple of
        * The symbol ``s`` used in the polynomial
        * The symbol ``t`` used in the polynomial
        * The triangle :math:`B(s, t)`.
    """
    nodes_sym = to_symbolic(nodes)

    s, t = sympy.symbols("s, t")
    b_polynomial = nodes_sym * triangle_weights(degree, s, t)
    b_polynomial.simplify()

    factored = [value.factor() for value in b_polynomial]
    return s, t, sympy.Matrix(factored).reshape(*b_polynomial.shape)


@require_sympy
def implicitize_3d(x_fn, y_fn, z_fn, s, t):
    """Implicitize a 3D parametric triangle.

    Args:
        x_fn (sympy.Expr): Function :math:`x(s)` in the triangle.
        y_fn (sympy.Expr): Function :math:`y(s)` in the triangle.
        z_fn (sympy.Expr): Function :math:`z(s)` in the triangle.
        s (sympy.Symbol): The first symbol used to define ``x_fn``, ``y_fn``
            and ``z_fn``.
        t (sympy.Symbol): The second symbol used to define ``x_fn``, ``y_fn``
            and ``z_fn``.

    Returns:
        sympy.Expr: The implicitized function :math:`f(x, y, z)` such that the
        triangle satisfies :math:`f(x(s, t), y(s, t), z(s, t)) = 0`.
    """
    x_sym, y_sym, z_sym = sympy.symbols("x, y, z")

    f_xy = sympy.resultant(x_fn - x_sym, y_fn - y_sym, s)
    f_yz = sympy.resultant(y_fn - x_sym, z_fn - z_sym, s)
    return sympy.resultant(f_xy, f_yz, t).factor()


@require_sympy
def implicitize_triangle(nodes, degree):
    """Implicitize a 3D parametric triangle, given the nodes.

    .. note::

       This function assumes (but does not check) that the caller has verified
       ``nodes`` represents a 3D triangle. If it **does not**, this function
       will error out when tuple-unpacking the polynomial.

    Args:
        nodes (numpy.ndarray): Nodes defining a B |eacute| zier triangle.
        degree (int): The degree of the triangle. This is assumed to
            correctly correspond to the number of ``nodes``.

    Returns:
        sympy.Expr: The implicitized function :math:`f(x, y, z)` such that the
        triangle satisfies :math:`f(x(s, t), y(s, t), z(s, t)) = 0`.
    """
    s, t, b_polynomial = triangle_as_polynomial(nodes, degree)
    x_fn, y_fn, z_fn = b_polynomial
    return implicitize_3d(x_fn, y_fn, z_fn, s, t)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pure Python helper methods for intersecting B |eacute| zier shapes.

.. |eacute| unicode:: U+000E9 .. LATIN SMALL LETTER E WITH ACUTE
   :trim:
"""

import enum

import numpy as np

from bezier import _py_curve_helpers
from bezier import _py_helpers


# For ``full_newton()``.
ZERO_THRESHOLD = 0.5 ** 10  # ~1e-3
MAX_NEWTON_ITERATIONS = 10
NEWTON_ERROR_RATIO = 0.5 ** 36
NEWTON_NO_CONVERGE = """\
Unsupported multiplicity.

Newton's method failed to converge to a solution under the
following assumptions:

- The starting ``s-t`` values were already near a solution
- The root / solution has multiplicity 1 or 2
  - 1: The root is "simple", i.e. the curves are not tangent
       and have no self-intersections at the point of intersection.
  - 2: The root is a double root, i.e. the curves are tangent
       but have different curvatures at the point of intersection.

The failure to converge may have been caused by one of:

- The root was of multiplicity greater than 2
- The curves don't actually intersect, though they come very close
- Numerical issues caused the iteration to leave the region
  of convergence
"""


def newton_refine(s, nodes1, t, nodes2):
    r"""Apply one step of 2D Newton's method.

    .. note::

       There is also a Fortran implementation of this function, which
       will be used if it can be built.

    We want to use Newton's method on the function

    .. math::

       F(s, t) = B_1(s) - B_2(t)

    to refine :math:`\left(s_{\ast}, t_{\ast}\right)`. Using this,
    and the Jacobian :math:`DF`, we "solve"

    .. math::

       \left[\begin{array}{c}
           0 \\ 0 \end{array}\right] \approx
           F\left(s_{\ast} + \Delta s, t_{\ast} + \Delta t\right) \approx
           F\left(s_{\ast}, t_{\ast}\right) +
           \left[\begin{array}{c c}
               B_1'\left(s_{\ast}\right) &
               - B_2'\left(t_{\ast}\right) \end{array}\right]
           \left[\begin{array}{c}
               \Delta s \\ \Delta t \end{array}\right]

    and refine with the component updates :math:`\Delta s` and
    :math:`\Delta t`.

    .. note::

       This implementation assumes the curves live in
       :math:`\mathbf{R}^2`.

    For example, the curves

    .. math::

        \begin{align*}
        B_1(s) &= \left[\begin{array}{c} 0 \\ 0 \end{array}\right] (1 - s)^2
            + \left[\begin{array}{c} 2 \\ 4 \end{array}\right] 2s(1 - s)
            + \left[\begin{array}{c} 4 \\ 0 \end{array}\right] s^2 \\
        B_2(t) &= \left[\begin{array}{c} 2 \\ 0 \end{array}\right] (1 - t)
            + \left[\begin{array}{c} 0 \\ 3 \end{array}\right] t
        \end{align*}

    intersect at the point
    :math:`B_1\left(\frac{1}{4}\right) = B_2\left(\frac{1}{2}\right) =
    \frac{1}{2} \left[\begin{array}{c} 2 \\ 3 \end{array}\right]`.

    However, starting from the wrong point we have

    .. math::

        \begin{align*}
        F\left(\frac{3}{8}, \frac{1}{4}\right) &= \frac{1}{8}
            \left[\begin{array}{c} 0 \\ 9 \end{array}\right] \\
        DF\left(\frac{3}{8}, \frac{1}{4}\right) &=
            \left[\begin{array}{c c}
            4 & 2 \\ 2 & -3 \end{array}\right] \\
        \Longrightarrow \left[\begin{array}{c} \Delta s \\ \Delta t
            \end{array}\right] &= \frac{9}{64} \left[\begin{array}{c}
            -1 \\ 2 \end{array}\right].
        \end{align*}

    .. image:: ../images/newton_refine1.png
       :align: center

    .. testsetup:: newton-refine1, newton-refine2, newton-refine3

       import numpy as np
       import bezier
       from bezier._py_intersection_helpers import newton_refine

       machine_eps = np.finfo(np.float64).eps

       def realroots(*coeffs):
           all_roots = np.roots(coeffs)
           return all_roots[np.where(all_roots.imag == 0.0)].real

    .. doctest:: newton-refine1

       >>> nodes1 = np.asfortranarray([
       ...     [0.0, 2.0, 4.0],
       ...     [0.0, 4.0, 0.0],
       ... ])
       >>> nodes2 = np.asfortranarray([
       ...     [2.0, 0.0],
       ...     [0.0, 3.0],
       ... ])
       >>> s, t = 0.375, 0.25
       >>> new_s, new_t = newton_refine(s, nodes1, t, nodes2)
       >>> 64.0 * (new_s - s)
       -9.0
       >>> 64.0 * (new_t - t)
       18.0

    .. testcleanup:: newton-refine1

       import make_images
       curve1 = bezier.Curve(nodes1, degree=2)
       curve2 = bezier.Curve(nodes2, degree=1)
       make_images.newton_refine1(s, new_s, curve1, t, new_t, curve2)

    For "typical" curves, we converge to a solution quadratically.
    This means that the number of correct digits doubles every
    iteration (until machine precision is reached).

    .. image:: ../images/newton_refine2.png
       :align: center

    .. doctest:: newton-refine2

       >>> nodes1 = np.asfortranarray([
       ...     [0.0, 0.25,  0.5, 0.75, 1.0],
       ...     [0.0, 2.0 , -2.0, 2.0 , 0.0],
       ... ])
       >>> nodes2 = np.asfortranarray([
       ...     [0.0, 0.25, 0.5, 0.75, 1.0],
       ...     [1.0, 0.5 , 0.5, 0.5 , 0.0],
       ... ])
       >>> # The expected intersection is the only real root of
       >>> # 28 s^3 - 30 s^2 + 9 s - 1.
       >>> expected, = realroots(28, -30, 9, -1)
       >>> s_vals = [0.625, None, None, None, None]
       >>> t = 0.625
       >>> np.log2(abs(expected - s_vals[0]))
       -4.399...
       >>> s_vals[1], t = newton_refine(s_vals[0], nodes1, t, nodes2)
       >>> np.log2(abs(expected - s_vals[1]))
       -7.901...
       >>> s_vals[2], t = newton_refine(s_vals[1], nodes1, t, nodes2)
       >>> np.log2(abs(expected - s_vals[2]))
       -16.010...
       >>> s_vals[3], t = newton_refine(s_vals[2], nodes1, t, nodes2)
       >>> np.log2(abs(expected - s_vals[3]))
       -32.110...
       >>> s_vals[4], t = newton_refine(s_vals[3], nodes1, t, nodes2)
       >>> np.allclose(s_vals[4], expected, rtol=6 * machine_eps, atol=0.0)
       True

    .. testcleanup:: newton-refine2

       import make_images
       curve1 = bezier.Curve(nodes1, degree=4)
       curve2 = bezier.Curve(nodes2, degree=4)
       make_images.newton_refine2(s_vals, curve1, curve2)

    However, when the intersection occurs at a point of tangency,
    the convergence becomes linear. This means that the number of
    correct digits added each iteration is roughly constant.

    .. image:: ../images/newton_refine3.png
       :align: center

    .. doctest:: newton-refine3

       >>> nodes1 = np.asfortranarray([
       ...     [0.0, 0.5, 1.0],
       ...     [0.0, 1.0, 0.0],
       ... ])
       >>> nodes2 = np.asfortranarray([
       ...     [0.0, 1.0],
       ...     [0.5, 0.5],
       ... ])
       >>> expected = 0.5
       >>> s_vals = [0.375, None, None, None, None, None]
       >>> t = 0.375
       >>> np.log2(abs(expected - s_vals[0]))
       -3.0
       >>> s_vals[1], t = newton_refine(s_vals[0], nodes1, t, nodes2)
       >>> np.log2(abs(expected - s_vals[1]))
       -4.0
       >>> s_vals[2], t = newton_refine(s_vals[1], nodes1, t, nodes2)
       >>> np.log2(abs(expected - s_vals[2]))
       -5.0
       >>> s_vals[3], t = newton_refine(s_vals[2], nodes1, t, nodes2)
       >>> np.log2(abs(expected - s_vals[3]))
       -6.0
       >>> s_vals[4], t = newton_refine(s_vals[3], nodes1, t, nodes2)
       >>> np.log2(abs(expected - s_vals[4]))
       -7.0
       >>> s_vals[5], t = newton_refine(s_vals[4], nodes1, t, nodes2)
       >>> np.log2(abs(expected - s_vals[5]))
       -8.0

    .. testcleanup:: newton-refine3

       import make_images
       curve1 = bezier.Curve(nodes1, degree=2)
       curve2 = bezier.Curve(nodes2, degree=1)
       make_images.newton_refine3(s_vals, curve1, curve2)

    Unfortunately, the process terminates with an error that is not close
    to machine precision :math:`\varepsilon` when
    :math:`\Delta s = \Delta t = 0`.

    .. testsetup:: newton-refine3-continued

       import numpy as np
       import bezier
       from bezier._py_intersection_helpers import newton_refine

       nodes1 = np.asfortranarray([
           [0.0, 0.5, 1.0],
           [0.0, 1.0, 0.0],
       ])
       nodes2 = np.asfortranarray([
           [0.0, 1.0],
           [0.5, 0.5],
       ])

    .. doctest:: newton-refine3-continued

       >>> s1 = t1 = 0.5 - 0.5**27
       >>> np.log2(0.5 - s1)
       -27.0
       >>> s2, t2 = newton_refine(s1, nodes1, t1, nodes2)
       >>> s2 == t2
       True
       >>> np.log2(0.5 - s2)
       -28.0
       >>> s3, t3 = newton_refine(s2, nodes1, t2, nodes2)
       >>> s3 == t3 == s2
       True

    Due to round-off near the point of tangency, the final error
    resembles :math:`\sqrt{\varepsilon}` rather than machine
    precision as expected.

    .. note::

       The following is not implemented in this function. It's just
       an exploration on how the shortcomings might be addressed.

    However, this can be overcome. At the point of tangency, we want
    :math:`B_1'(s) \parallel B_2'(t)`. This can be checked numerically via

    .. math::

        B_1'(s) \times B_2'(t) = 0.

    For the last example (the one that converges linearly), this is

    .. math::

        0 = \left[\begin{array}{c} 1 \\ 2 - 4s \end{array}\right] \times
            \left[\begin{array}{c} 1 \\ 0 \end{array}\right] = 4 s - 2.

    With this, we can modify Newton's method to find a zero of the
    over-determined system

    .. math::

        G(s, t) = \left[\begin{array}{c} B_0(s) - B_1(t) \\
            B_1'(s) \times B_2'(t) \end{array}\right] =
            \left[\begin{array}{c} s - t \\ 2 s (1 - s) - \frac{1}{2} \\
            4 s - 2\end{array}\right].

    Since :math:`DG` is :math:`3 \times 2`, we can't invert it. However,
    we can find a least-squares solution:

    .. math::

        \left(DG^T DG\right) \left[\begin{array}{c}
            \Delta s \\ \Delta t \end{array}\right] = -DG^T G.

    This only works if :math:`DG` has full rank. In this case, it does
    since the submatrix containing the first and last rows has rank two:

    .. math::

        DG = \left[\begin{array}{c c} 1 & -1 \\
            2 - 4 s & 0 \\
            4 & 0 \end{array}\right].

    Though this avoids a singular system, the normal equations have a
    condition number that is the square of the condition number of the matrix.

    Starting from :math:`s = t = \frac{3}{8}` as above:

    .. testsetup:: newton-refine4

       import numpy as np
       from bezier import _py_helpers

       def modified_update(s, t):
           minus_G = np.asfortranarray([
               [t - s],
               [0.5 - 2.0 * s * (1.0 - s)],
               [2.0 - 4.0 * s],
           ])
           DG = np.asfortranarray([
               [1.0, -1.0],
               [2.0 - 4.0 * s, 0.0],
               [4.0, 0.0],
           ])
           DG_t = np.asfortranarray(DG.T)

           LHS = _py_helpers.matrix_product(DG_t, DG)
           RHS = _py_helpers.matrix_product(DG_t, minus_G)
           delta_params = np.linalg.solve(LHS, RHS)
           delta_s, delta_t = delta_params.flatten()
           return s + delta_s, t + delta_t

    .. doctest:: newton-refine4

       >>> s0, t0 = 0.375, 0.375
       >>> np.log2(0.5 - s0)
       -3.0
       >>> s1, t1 = modified_update(s0, t0)
       >>> s1 == t1
       True
       >>> 1040.0 * s1
       519.0
       >>> np.log2(0.5 - s1)
       -10.022...
       >>> s2, t2 = modified_update(s1, t1)
       >>> s2 == t2
       True
       >>> np.log2(0.5 - s2)
       -31.067...
       >>> s3, t3 = modified_update(s2, t2)
       >>> s3 == t3 == 0.5
       True

    Args:
        s (float): Parameter of a near-intersection along the first curve.
        nodes1 (numpy.ndarray): Nodes of first curve forming intersection.
        t (float): Parameter of a near-intersection along the second curve.
        nodes2 (numpy.ndarray): Nodes of second curve forming intersection.

    Returns:
        Tuple[float, float]: The refined parameters from a single Newton
        step.

    Raises:
        ValueError: If the Jacobian is singular at ``(s, t)``.
    """
    # NOTE: We form -F(s, t) since we want to solve -DF^{-1} F(s, t).
    func_val = _py_curve_helpers.evaluate_multi(
        nodes2, np.asfortranarray([t])
    ) - _py_curve_helpers.evaluate_multi(nodes1, np.asfortranarray([s]))
    if np.all(func_val == 0.0):
        # No refinement is needed.
        return s, t

    # NOTE: This assumes the curves are 2D.
    jac_mat = np.empty((2, 2), order="F")
    jac_mat[:, :1] = _py_curve_helpers.evaluate_hodograph(s, nodes1)
    jac_mat[:, 1:] = -_py_curve_helpers.evaluate_hodograph(t, nodes2)
    # Solve the system.
    singular, delta_s, delta_t = _py_helpers.solve2x2(jac_mat, func_val[:, 0])
    if singular:
        raise ValueError("Jacobian is singular.")

    return s + delta_s, t + delta_t


class NewtonSimpleRoot:  # pylint: disable=too-few-public-methods
    r"""Callable object that facilitates Newton's method.

    This is meant to be used to compute the Newton update via:

    .. math::

       DF(s, t) \left[\begin{array}{c}
           \Delta s \\ \Delta t \end{array}\right] = -F(s, t).

    Args:
        nodes1 (numpy.ndarray): Control points of the first curve.
        first_deriv1 (numpy.ndarray): Control points of the curve
            :math:`B_1'(s)`.
        nodes2 (numpy.ndarray): Control points of the second curve.
        first_deriv2 (numpy.ndarray): Control points of the curve
            :math:`B_2'(t)`.
    """

    def __init__(self, nodes1, first_deriv1, nodes2, first_deriv2):
        self.nodes1 = nodes1
        self.first_deriv1 = first_deriv1
        self.nodes2 = nodes2
        self.first_deriv2 = first_deriv2

    def __call__(self, s, t):
        r"""This computes :math:`F = B_1(s) - B_2(t)` and :math:`DF(s, t)`.

        .. note::

           There is **almost** identical code in :func:`.newton_refine`, but
           that code can avoid computing the ``first_deriv1`` and
           ``first_deriv2`` nodes in cases that :math:`F(s, t) = 0` whereas
           this function assumes they have been given.

        In the case that :math:`DF(s, t)` is singular, the assumption is that
        the intersection has a multiplicity higher than one (i.e. the root is
        non-simple). **Near** a simple root, it must be the case that
        :math:`DF(s, t)` has non-zero determinant, so due to continuity, we
        assume the Jacobian will be invertible nearby.

        Args:
            s (float): The parameter where we'll compute :math:`B_1(s)` and
                :math:`DF(s, t)`.
            t (float): The parameter where we'll compute :math:`B_2(t)` and
                :math:`DF(s, t)`.

        Returns:
            Tuple[Optional[numpy.ndarray], numpy.ndarray]: Pair of

            * The LHS matrix ``DF``, a ``2 x 2`` array. If ``F == 0`` then
              this matrix won't be computed and :data:`None` will be returned.
            * The RHS vector ``F``, a ``2 x 1`` array.
        """
        s_vals = np.asfortranarray([s])
        b1_s = _py_curve_helpers.evaluate_multi(self.nodes1, s_vals)
        t_vals = np.asfortranarray([t])
        b2_t = _py_curve_helpers.evaluate_multi(self.nodes2, t_vals)
        func_val = b1_s - b2_t
        if np.all(func_val == 0.0):
            return None, func_val

        else:
            jacobian = np.empty((2, 2), order="F")
            jacobian[:, :1] = _py_curve_helpers.evaluate_multi(
                self.first_deriv1, s_vals
            )
            jacobian[:, 1:] = -_py_curve_helpers.evaluate_multi(
                self.first_deriv2, t_vals
            )
            return jacobian, func_val


class NewtonDoubleRoot:  # pylint: disable=too-few-public-methods
    r"""Callable object that facilitates Newton's method for double roots.

    This is an augmented version of :class:`NewtonSimpleRoot`.

    For non-simple intersections (i.e. multiplicity greater than 1),
    the curves will be tangent, which forces :math:`B_1'(s) \times B_2'(t)`
    to be zero. Unfortunately, that quantity is also equal to the
    determinant of the Jacobian, so :math:`DF` will not be full rank.

    In order to produce a system that **can** be solved, an
    an augmented function is computed:

    .. math::

        G(s, t) = \left[\begin{array}{c}
            F(s, t) \\ \hline
            B_1'(s) \times B_2'(t)
            \end{array}\right]

    The use of :math:`B_1'(s) \times B_2'(t)` (with lowered degree in
    :math:`s` and :math:`t`) means that the rank deficiency in
    :math:`DF` can be fixed in:

    .. math::

        DG(s, t) = \left[\begin{array}{c | c}
            B_1'(s) & -B_2'(t) \\ \hline
            B_1''(s) \times B_2'(t) & B_1'(s) \times B_2''(t)
            \end{array}\right]

    (This may not always be full rank, but in the double root / multiplicity
    2 case it will be full rank near a solution.)

    Rather than finding a least squares solution to the overdetermined system

    .. math::

       DG(s, t) \left[\begin{array}{c}
           \Delta s \\ \Delta t \end{array}\right] = -G(s, t)

    we find a solution to the square (and hopefully full rank) system:

    .. math::

       DG^T DG \left[\begin{array}{c}
           \Delta s \\ \Delta t \end{array}\right] = -DG^T G.

    Forming ``DG^T DG`` squares the condition number, so it would be "better"
    to use :func:`~numpy.linalg.lstsq` (which wraps the LAPACK routine
    ``dgelsd``). However, using :func:`.solve2x2` is **much** more
    straightforward and in practice this is just as accurate.

    Args:
        nodes1 (numpy.ndarray): Control points of the first curve.
        first_deriv1 (numpy.ndarray): Control points of the curve
            :math:`B_1'(s)`.
        second_deriv1 (numpy.ndarray): Control points of the curve
            :math:`B_1''(s)`.
        nodes2 (numpy.ndarray): Control points of the second curve.
        first_deriv2 (numpy.ndarray): Control points of the curve
            :math:`B_2'(t)`.
        second_deriv2 (numpy.ndarray): Control points of the curve
            :math:`B_2''(t)`.
    """

    def __init__(
        self,
        nodes1,
        first_deriv1,
        second_deriv1,
        nodes2,
        first_deriv2,
        second_deriv2,
    ):
        self.nodes1 = nodes1
        self.first_deriv1 = first_deriv1
        self.second_deriv1 = second_deriv1
        self.nodes2 = nodes2
        self.first_deriv2 = first_deriv2
        self.second_deriv2 = second_deriv2

    def __call__(self, s, t):
        r"""This computes :math:`DG^T G` and :math:`DG^T DG`.

        If :math:`DG^T DG` is not full rank, this means either :math:`DG`
        was not full rank or that it was, but with a relatively high condition
        number. So, in the case that :math:`DG^T DG` is singular, the
        assumption is that the intersection has a multiplicity higher than two.

        Args:
            s (float): The parameter where we'll compute :math:`G(s, t)` and
                :math:`DG(s, t)`.
            t (float): The parameter where we'll compute :math:`G(s, t)` and
                :math:`DG(s, t)`.

        Returns:
            Tuple[Optional[numpy.ndarray], Optional[numpy.ndarray]]: Pair of

            * The LHS matrix ``DG^T DG``, a ``2 x 2`` array. If ``G == 0`` then
              this matrix won't be computed and :data:`None` will be returned.
            * The RHS vector ``DG^T G``, a ``2 x 1`` array.
        """
        s_vals = np.asfortranarray([s])
        b1_s = _py_curve_helpers.evaluate_multi(self.nodes1, s_vals)
        b1_ds = _py_curve_helpers.evaluate_multi(self.first_deriv1, s_vals)
        t_vals = np.asfortranarray([t])
        b2_t = _py_curve_helpers.evaluate_multi(self.nodes2, t_vals)
        b2_dt = _py_curve_helpers.evaluate_multi(self.first_deriv2, t_vals)
        func_val = np.empty((3, 1), order="F")
        func_val[:2, :] = b1_s - b2_t
        func_val[2, :] = _py_helpers.cross_product(b1_ds[:, 0], b2_dt[:, 0])
        if np.all(func_val == 0.0):
            return None, func_val[:2, :]

        else:
            jacobian = np.empty((3, 2), order="F")
            jacobian[:2, :1] = b1_ds
            jacobian[:2, 1:] = -b2_dt
            if self.second_deriv1.size == 0:
                jacobian[2, 0] = 0.0
            else:
                jacobian[2, 0] = _py_helpers.cross_product(
                    _py_curve_helpers.evaluate_multi(
                        self.second_deriv1, s_vals
                    )[:, 0],
                    b2_dt[:, 0],
                )
            if self.second_deriv2.size == 0:
                jacobian[2, 1] = 0.0
            else:
                jacobian[2, 1] = _py_helpers.cross_product(
                    b1_ds[:, 0],
                    _py_curve_helpers.evaluate_multi(
                        self.second_deriv2, t_vals
                    )[:, 0],
                )
            modified_lhs = _py_helpers.matrix_product(jacobian.T, jacobian)
            modified_rhs = _py_helpers.matrix_product(jacobian.T, func_val)
            return modified_lhs, modified_rhs


def newton_iterate(evaluate_fn, s, t):
    r"""Perform a Newton iteration.

    In this function, we assume that :math:`s` and :math:`t` are nonzero,
    this makes convergence easier to detect since "relative error" at
    ``0.0`` is not a useful measure.

    There are several tolerance / threshold quantities used below:

    * :math:`10` (:attr:`MAX_NEWTON_ITERATIONS`) iterations will be done before
      "giving up". This is based on the assumption that we are already starting
      near a root, so quadratic convergence should terminate quickly.
    * :math:`\tau = \frac{1}{4}` is used as the boundary between linear
      and superlinear convergence. So if the current error
      :math:`\|p_{n + 1} - p_n\|` is not smaller than :math:`\tau` times
      the previous error :math:`\|p_n - p_{n - 1}\|`, then convergence
      is considered to be linear at that point.
    * :math:`\frac{2}{3}` of all iterations must be converging linearly
      for convergence to be stopped (and moved to the next regime). This
      will only be checked after 4 or more updates have occurred.
    * :math:`\tau = 2^{-42}` (:attr:`NEWTON_ERROR_RATIO`) is used to
      determine that an update is sufficiently small to stop iterating. So if
      the error :math:`\|p_{n + 1} - p_n\|` smaller than :math:`\tau` times
      size of the term being updated :math:`\|p_n\|`, then we
      exit with the "correct" answer.

    It is assumed that ``evaluate_fn`` will use a Jacobian return value of
    :data:`None` to indicate that :math:`F(s, t)` is exactly ``0.0``. We
    **assume** that if the function evaluates to exactly ``0.0``, then we are
    at a solution. It is possible however, that badly parameterized curves
    can evaluate to exactly ``0.0`` for inputs that are relatively far away
    from a solution (see issue #21).

    Args:
        evaluate_fn (Callable[Tuple[float, float], tuple]): A callable
            which takes :math:`s` and :math:`t` and produces an evaluated
            function value and the Jacobian matrix.
        s (float): The (first) parameter where the iteration will start.
        t (float): The (second) parameter where the iteration will start.

    Returns:
        Tuple[bool, float, float]: The triple of

        * Flag indicating if the iteration converged.
        * The current :math:`s` value when the iteration stopped.
        * The current :math:`t` value when the iteration stopped.
    """
    # Several quantities will be tracked throughout the iteration:
    # * norm_update_prev: ||p{n}   - p{n-1}|| = ||dp{n-1}||
    # * norm_update     : ||p{n+1} - p{n}  || = ||dp{n}  ||
    # * linear_updates  : This is a count on the number of times that
    #                     ``dp{n}`` "looks like" ``dp{n-1}`` (i.e.
    #                     is within a constant factor of it).
    norm_update_prev = None
    norm_update = None
    linear_updates = 0  # Track the number of "linear" updates.
    current_s = s
    current_t = t
    for index in range(MAX_NEWTON_ITERATIONS):
        jacobian, func_val = evaluate_fn(current_s, current_t)
        if jacobian is None:
            return True, current_s, current_t

        singular, delta_s, delta_t = _py_helpers.solve2x2(
            jacobian, func_val[:, 0]
        )
        if singular:
            break

        norm_update_prev = norm_update
        norm_update = np.linalg.norm([delta_s, delta_t], ord=2)
        # If ||p{n} - p{n-1}|| > 0.25 ||p{n-1} - p{n-2}||, then that means
        # our convergence is acting linear at the current step.
        if index > 0 and norm_update > 0.25 * norm_update_prev:
            linear_updates += 1
        # If ``>=2/3`` of the updates have been linear, we are near a
        # non-simple root. (Make sure at least 5 updates have occurred.)
        if index >= 4 and 3 * linear_updates >= 2 * index:
            break

        # Determine the norm of the "old" solution before updating.
        norm_soln = np.linalg.norm([current_s, current_t], ord=2)
        current_s -= delta_s
        current_t -= delta_t
        if norm_update < NEWTON_ERROR_RATIO * norm_soln:
            return True, current_s, current_t

    return False, current_s, current_t


def full_newton_nonzero(s, nodes1, t, nodes2):
    r"""Perform a Newton iteration until convergence to a solution.

    This is the "implementation" for :func:`full_newton`. In this
    function, we assume that :math:`s` and :math:`t` are nonzero.

    Args:
        s (float): The parameter along the first curve where the iteration
            will start.
        nodes1 (numpy.ndarray): Control points of the first curve.
        t (float): The parameter along the second curve where the iteration
            will start.
        nodes2 (numpy.ndarray): Control points of the second curve.

    Returns:
        Tuple[float, float]: The pair of :math:`s` and :math:`t` values that
        Newton's method converged to.

    Raises:
        NotImplementedError: If Newton's method doesn't converge in either the
            multiplicity 1 or 2 cases.
    """
    # NOTE: We somewhat replicate code in ``evaluate_hodograph()``
    #       here. This is so we don't re-compute the nodes for the first
    #       (and possibly second) derivatives every time they are evaluated.
    _, num_nodes1 = np.shape(nodes1)
    first_deriv1 = (num_nodes1 - 1) * (nodes1[:, 1:] - nodes1[:, :-1])
    _, num_nodes2 = np.shape(nodes2)
    first_deriv2 = (num_nodes2 - 1) * (nodes2[:, 1:] - nodes2[:, :-1])
    evaluate_fn = NewtonSimpleRoot(nodes1, first_deriv1, nodes2, first_deriv2)
    converged, current_s, current_t = newton_iterate(evaluate_fn, s, t)
    if converged:
        return current_s, current_t

    # If Newton's method did not converge, then assume the root is not simple.
    second_deriv1 = (num_nodes1 - 2) * (
        first_deriv1[:, 1:] - first_deriv1[:, :-1]
    )
    second_deriv2 = (num_nodes2 - 2) * (
        first_deriv2[:, 1:] - first_deriv2[:, :-1]
    )
    evaluate_fn = NewtonDoubleRoot(
        nodes1,
        first_deriv1,
        second_deriv1,
        nodes2,
        first_deriv2,
        second_deriv2,
    )
    converged, current_s, current_t = newton_iterate(
        evaluate_fn, current_s, current_t
    )
    if converged:
        return current_s, current_t

    raise NotImplementedError(NEWTON_NO_CONVERGE)


def full_newton(s, nodes1, t, nodes2):
    r"""Perform a Newton iteration until convergence to a solution.

    This assumes :math:`s` and :math:`t` are sufficiently close to an
    intersection. It **does not** govern the maximum distance away
    that the solution can lie, though the subdivided intervals that contain
    :math:`s` and :math:`t` could be used.

    To avoid round-off issues near ``0.0``, this reverses the direction
    of a curve and replaces the parameter value :math:`\nu` with
    :math:`1 - \nu` whenever :math:`\nu < \tau` (here we use a threshold
    :math:`\tau` equal to :math:`2^{-10}`, i.e. ``ZERO_THRESHOLD``).

    Args:
        s (float): The parameter along the first curve where the iteration
            will start.
        nodes1 (numpy.ndarray): Control points of the first curve.
        t (float): The parameter along the second curve where the iteration
            will start.
        nodes2 (numpy.ndarray): Control points of the second curve.

    Returns:
        Tuple[float, float]: The pair of :math:`s` and :math:`t` values that
        Newton's method converged to.
    """
    if s < ZERO_THRESHOLD:
        reversed1 = np.asfortranarray(nodes1[:, ::-1])
        if t < ZERO_THRESHOLD:
            reversed2 = np.asfortranarray(nodes2[:, ::-1])
            refined_s, refined_t = full_newton_nonzero(
                1.0 - s, reversed1, 1.0 - t, reversed2
            )
            return 1.0 - refined_s, 1.0 - refined_t

        else:
            refined_s, refined_t = full_newton_nonzero(
                1.0 - s, reversed1, t, nodes2
            )
            return 1.0 - refined_s, refined_t

    else:
        if t < ZERO_THRESHOLD:
            reversed2 = np.asfortranarray(nodes2[:, ::-1])
            refined_s, refined_t = full_newton_nonzero(
                s, nodes1, 1.0 - t, reversed2
            )
            return refined_s, 1.0 - refined_t

        else:
            return full_newton_nonzero(s, nodes1, t, nodes2)


class IntersectionClassification(enum.Enum):
    """Enum classifying the "interior" curve in an intersection.

    Provided as the output values for :func:`.classify_intersection`.
    """

    FIRST = 0
    """The first curve is on the interior."""
    SECOND = 1
    """The second curve is on the interior."""
    OPPOSED = 2
    """Tangent intersection with opposed interiors."""
    TANGENT_FIRST = 3
    """Tangent intersection, first curve is on the interior."""
    TANGENT_SECOND = 4
    """Tangent intersection, second curve is on the interior."""
    IGNORED_CORNER = 5
    """Intersection at a corner, interiors don't intersect."""
    TANGENT_BOTH = 6
    """Tangent intersection, both curves are interior from some perspective."""
    COINCIDENT = 7
    """Intersection is actually an endpoint of a coincident segment."""
    COINCIDENT_UNUSED = 8
    """Unused because the edges are moving in opposite directions."""


class Intersection:  # pylint: disable=too-few-public-methods
    """Representation of a curve-curve intersection.

    Args:
        index_first (int): The index of the first curve within a list of
            curves. Expected to be used to index within the three edges of
            a triangle.
        s (float): The parameter along the first curve where the
            intersection occurs.
        index_second (int): The index of the second curve within a list of
            curves. Expected to be used to index within the three edges of
            a triangle.
        t (float): The parameter along the second curve where the
            intersection occurs.
        interior_curve (Optional[ \
            ~bezier._py_intersection_helpers.IntersectionClassification]): The
            classification of the intersection.
    """

    __slots__ = ("index_first", "s", "index_second", "t", "interior_curve")

    def __init__(self, index_first, s, index_second, t, interior_curve=None):
        self.index_first = index_first
        """int: Index of the first curve within a list of edges."""
        self.s = s
        """float: The intersection parameter for the first curve."""
        self.index_second = index_second
        """int: Index of the second curve within a list of edges."""
        self.t = t
        """float: The intersection parameter for the second curve."""
        self.interior_curve = interior_curve
        """IntersectionClassification: Which of the curves is on the interior.

        See :func:`.classify_intersection` for more details.
        """

    @property
    def __dict__(self):
        """dict: Dictionary of current intersection's property namespace.

        This is just a stand-in property for the usual ``__dict__``. This
        class defines ``__slots__`` so by default would not provide a
        ``__dict__``.

        This also means that the current object can't be modified by the
        returned dictionary.
        """
        return {
            "index_first": self.index_first,
            "s": self.s,
            "index_second": self.index_second,
            "t": self.t,
            "interior_curve": self.interior_curve,
        }


class IntersectionStrategy(enum.Enum):
    """Enum determining the type of intersection algorithm to use."""

    GEOMETRIC = 0
    """Geometric approach to intersection (via subdivision)."""
    ALGEBRAIC = 1
    """Algebraic approach to intersection (via implicitization)."""



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helpers for ``setup.py`` specific to Windows."""

import os
import shutil
import sys

import setup_helpers

LIB_DIR = os.path.join("bezier", "lib")
DLL_DIR = os.path.join("bezier", "extra-dll")
DLL_NAME = "bezier.dll"
DEF_NAME = "bezier.def"
LIB_NAME = "bezier.lib"
# See: https://docs.python.org/3/library/platform.html#cross-platform
if sys.maxsize == 2 ** 63 - 1:
    MACHINE_TYPE = "/machine:x64"
elif sys.maxsize == 2 ** 31 - 1:
    MACHINE_TYPE = "/machine:x86"
else:  # pragma: NO COVER
    raise ImportError("Unexpected maxsize", sys.maxsize)


def _ensure_exists(*dir_names):
    for dir_name in dir_names:
        if not os.path.exists(dir_name):
            os.makedirs(dir_name)


def make_static_lib(build_ext_cmd, obj_files):
    f90_compiler = build_ext_cmd.F90_COMPILER
    c_compiler = f90_compiler.c_compiler
    gfortran_exe = f90_compiler.compiler_f90[0]
    static_lib_dir = os.path.join(build_ext_cmd.build_lib, LIB_DIR)
    extra_dll_dir = os.path.join(build_ext_cmd.build_lib, DLL_DIR)
    _ensure_exists(static_lib_dir, extra_dll_dir, build_ext_cmd.build_temp)
    # Add build lib dir to extensions (for linking).
    for extension in build_ext_cmd.extensions:
        library_dirs = extension.library_dirs
        if static_lib_dir not in library_dirs:
            library_dirs.append(static_lib_dir)
    # NOTE: Would prefer to use a `tuple` but command must be
    #       mutable on Windows.
    temp_dll_filename = os.path.join(build_ext_cmd.build_temp, DLL_NAME)
    cmd = [gfortran_exe, "-static", "-shared", "-o", temp_dll_filename]
    cmd.extend(obj_files)
    def_filename = os.path.join(build_ext_cmd.build_temp, DEF_NAME)
    cmd.append("-Wl,--output-def," + def_filename)
    f90_compiler.spawn(cmd)
    # NOTE: This assumes, but does not check that ``c_compiler.initialized``
    #       is True.
    temp_lib_filename = os.path.join(build_ext_cmd.build_temp, LIB_NAME)
    cmd = [
        c_compiler.lib,
        "/def:" + def_filename,
        "/out:" + temp_lib_filename,
        MACHINE_TYPE,
    ]
    f90_compiler.spawn(cmd)
    # Move files for distribution from build temp. dir to build lib dir.
    dll_filename = os.path.join(extra_dll_dir, DLL_NAME)
    shutil.move(temp_dll_filename, dll_filename)
    lib_filename = os.path.join(static_lib_dir, LIB_NAME)
    shutil.move(temp_lib_filename, lib_filename)


def run_cleanup(build_ext_cmd):
    """Cleanup after ``BuildFortranThenExt.run``.

    For in-place builds, moves the built shared library into the source
    directory.
    """
    if not build_ext_cmd.inplace:
        return

    bezier_dir = os.path.join("src", "python", "bezier")
    shutil.move(os.path.join(build_ext_cmd.build_lib, LIB_DIR), bezier_dir)
    shutil.move(os.path.join(build_ext_cmd.build_lib, DLL_DIR), bezier_dir)


def _remove_fpic(compiler_flags):
    """Remove ``-fPIC`` from a given set of compiler flags.

    Args:
        compiler_flags (List[str]): Existing flags associated with a compiler.

    Returns:
        List[str]: The modified list.
    """
    return [value for value in compiler_flags if value != setup_helpers.FPIC]


def patch_f90_compiler(f90_compiler):
    """Patch up ``f90_compiler.library_dirs``.

    Updates flags in ``gfortran`` and ignores other compilers. The only
    modification is the removal of ``-fPIC`` since it is not used on Windows
    and the build flags turn warnings into errors.

    Args:
        f90_compiler (numpy.distutils.fcompiler.FCompiler): A Fortran compiler
            instance.
    """
    # NOTE: NumPy may not be installed, but we don't want **this** module to
    #       cause an import failure.
    from numpy.distutils.fcompiler import gnu

    # Only Windows.
    if os.name != "nt":
        return

    # Only ``gfortran``.
    if not isinstance(f90_compiler, gnu.Gnu95FCompiler):
        return

    f90_compiler.compiler_f77[:] = _remove_fpic(f90_compiler.compiler_f77)
    f90_compiler.compiler_f90[:] = _remove_fpic(f90_compiler.compiler_f90)
    c_compiler = f90_compiler.c_compiler
    if c_compiler.compiler_type != "msvc":
        raise NotImplementedError(
            "MSVC is the only supported C compiler on Windows."
        )


def patch_cmd(cmd_class):
    # Only Windows.
    if os.name != "nt":
        return

    cmd_class.CUSTOM_STATIC_LIB = make_static_lib
    cmd_class.USE_SHARED_LIBRARY = True
    cmd_class.CLEANUP = run_cleanup
